/*
 * Copyright 2015-2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.concurrent.locks.ReentrantLock;

import com.aquenos.epics.jackie.common.exception.ConcurrentNotificationException;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessEventMask;
import com.aquenos.epics.jackie.common.util.FutureCompletionListener;
import com.aquenos.epics.jackie.common.util.ListenableFuture;
import com.aquenos.epics.jackie.common.util.ListenerLockPolicy;
import com.aquenos.epics.jackie.common.util.SimpleListenableFuture;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmAcknowledgementStatus;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyChar;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyDouble;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyEnum;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyFloat;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyLong;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyShort;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyString;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmSeverity;
import com.aquenos.epics.jackie.common.value.ChannelAccessClassName;
import com.aquenos.epics.jackie.common.value.ChannelAccessControlsChar;
import com.aquenos.epics.jackie.common.value.ChannelAccessControlsDouble;
import com.aquenos.epics.jackie.common.value.ChannelAccessControlsFloat;
import com.aquenos.epics.jackie.common.value.ChannelAccessControlsLong;
import com.aquenos.epics.jackie.common.value.ChannelAccessControlsShort;
import com.aquenos.epics.jackie.common.value.ChannelAccessGettableValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsEnum;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsOnlyChar;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsOnlyDouble;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsOnlyFloat;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsOnlyLong;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsOnlyShort;
import com.aquenos.epics.jackie.common.value.ChannelAccessPuttableValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyChar;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyDouble;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyEnum;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyFloat;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyLong;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyShort;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyString;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeChar;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeDouble;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeEnum;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeFloat;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeLong;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeShort;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeString;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueFactory;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * <p>
 * Abstract base class for {@link ChannelAccessChannel} implementations. This
 * base class takes care of handling the connection state listeners and of
 * delegating the various specializations of the get and monitor methods to the
 * one most generic version.
 * </p>
 * 
 * <p>
 * This class uses a {@link HashSet} to keep all
 * {@link ChannelAccessConnectionListener ChannelAccessConnectionListeners} that
 * are registered with this channel. Derived classes should call the
 * {@link #updateConnectionState(boolean)} method in order to update the
 * connection state and notify the listeners. Implementations may choose to
 * override the
 * {@link #notifyConnectionListener(ChannelAccessConnectionListener, boolean)}
 * method in order to customize the handling of notifications (e.g. catch
 * certain exceptions possibly thrown by listeners).
 * </p>
 * 
 * <p>
 * All invocations of the various get methods are delegated to the
 * {@link #get(ChannelAccessValueType, int)} method, which has to be implemented
 * by derived classes. All invocations of the various monitor methods are
 * delegated to the
 * {@link #monitor(ChannelAccessValueType, int, ChannelAccessEventMask)} method,
 * which has to be implemented by derived classes.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public abstract class AbstractChannelAccessChannel
        implements ChannelAccessChannel {

    /**
     * Registration of a connection listener. This is a separate class because
     * we have to store some flags together with the lock object.
     * 
     * @author Sebastian Marsching
     */
    private static class ListenerRegistration {

        public final ChannelAccessConnectionListener listener;
        public boolean active = true;
        public boolean firstNotificationSent = false;
        public final ReentrantLock lock = new ReentrantLock();

        public ListenerRegistration(ChannelAccessConnectionListener listener) {
            this.listener = listener;
        }

    }

    /**
     * <p>
     * Custom implementation of a {@link SimpleListenableFuture} that is used
     * for the
     * {@link AbstractChannelAccessChannel#waitForConnectionState(boolean)}
     * method.
     * </p>
     * 
     * <p>
     * This implementation makes two modifications:
     * </p>
     * 
     * <p>
     * First, it uses the
     * {@link AbstractChannelAccessChannel#notifyWaitForConnectionStateFutureListener(ListenableFuture, FutureCompletionListener)}
     * method for notifying its completion listeners. This way, derived classes
     * can implement the notification mechanism in a way that is safe regarding
     * exceptions thrown by a listener.
     * </p>
     * 
     * <p>
     * Second, this implementation keeps a strong reference to the channel
     * instance until the future has completed. This reference is kept for two
     * reasons: First, it is needed to delegate the notification of listeners to
     * the channel implementation. Second, it ensures that a channel instance is
     * not destroyed as long as there a futures waiting for its connection state
     * to change. If this happened, the futures would never complete and threads
     * waiting for them would be blocked indefinitely.
     * </p>
     * 
     * @author Sebastian Marsching
     */
    private static class WaitForConnectionStateListenableFuture
            extends SimpleListenableFuture<Boolean> {

        private AbstractChannelAccessChannel channel;

        public WaitForConnectionStateListenableFuture(
                AbstractChannelAccessChannel channel,
                ListenerLockPolicy listenerLockPolicy) {
            super(listenerLockPolicy);
            this.channel = channel;
        }

        @Override
        protected void notifyCompletionListeners() {
            try {
                super.notifyCompletionListeners();
            } finally {
                // After the listeners have been notified, we can release the
                // reference to the channel. We do not need it any longer and
                // the channel can now safely be destroyed (if there are no
                // other references to it).
                channel = null;
            }
        }

        @Override
        protected void notifyCompletionListener(
                FutureCompletionListener<? super Boolean> listener) {
            channel.notifyWaitForConnectionStateFutureListener(this, listener);
        }

    }

    private boolean connected;
    private boolean connectedOnce;

    // We store the listener registrations in a map instead of a set. This way,
    // we can efficiently remove a registration for a specific listener.
    private HashMap<ChannelAccessConnectionListener, ListenerRegistration> listeners = new HashMap<ChannelAccessConnectionListener, ListenerRegistration>();
    private LinkedList<WaitForConnectionStateListenableFuture> waitForConnectedFutures = new LinkedList<WaitForConnectionStateListenableFuture>();
    private LinkedList<WaitForConnectionStateListenableFuture> waitForDisconnectedFutures = new LinkedList<WaitForConnectionStateListenableFuture>();

    /**
     * Synchronization object that this class synchronizes on when retrieving or
     * modifying the set of registered listeners.
     */
    protected final Object connectionStateLock = new Object();

    /**
     * Listener lock-policy used by this channel. The lock-policy affects the
     * behavior of the
     * {@link #removeConnectionListener(ChannelAccessConnectionListener)} method
     * and the behavior of the futures returned by
     * {@link #waitForConnectionState(boolean)}. In addition to that, it might
     * affect the behavior of methods implemented by child classes. Typically,
     * it will affect the behavior of monitors created by the
     * {@link #monitor(ChannelAccessValueType, int, ChannelAccessEventMask)}
     * method and futures returned by the
     * {@link #get(ChannelAccessValueType, int)} and
     * {@link #put(ChannelAccessPuttableValue)} method.
     */
    protected final ListenerLockPolicy listenerLockPolicy;

    /**
     * Default constructor. This constructor calls
     * {@link #AbstractChannelAccessChannel(ListenerLockPolicy)} with an
     * argument of {@link ListenerLockPolicy#BLOCK}.
     * 
     * @deprecated Since 2.0.0. Use
     *             {@link #AbstractChannelAccessChannel(ListenerLockPolicy)}
     *             instead.
     */
    protected AbstractChannelAccessChannel() {
        this(ListenerLockPolicy.BLOCK);
    }

    /**
     * Initializes this channel with the specified listener lock-policy. This
     * constructor is supposed to be used by classes that inherit from this
     * class. It initializes the channel with the disconnected state and an
     * empty set of connection listeners.
     * 
     * @param listenerLockPolicy
     *            listener lock-policy used by this channel. The lister-lock
     *            policy is used by the
     *            {@link #removeConnectionListener(ChannelAccessConnectionListener)}
     *            method and by the futures returned from the
     *            {@link #waitForConnectionState(boolean)} method. Child classes
     *            might use the policy for additional purposes.
     * @since EPICS Jackie 2.0.0
     */
    protected AbstractChannelAccessChannel(
            ListenerLockPolicy listenerLockPolicy) {
        if (listenerLockPolicy == null) {
            throw new NullPointerException(
                    "The listener lock-policy must not be null.");
        }
        this.listenerLockPolicy = listenerLockPolicy;
    }

    /**
     * <p>
     * Fails the waiting connection state futures. This method must be called by
     * the child class when this channel is {@linkplain #destroy() destroyed}.
     * It is important that it is called after the destroyed flag has been set
     * (and thus {@link #isDestroyed()} returns <code>true</code>) because only
     * this will ensure that no more futures will be added to the list of
     * waiting futures.
     * </p>
     * 
     * <p>
     * All waiting connection-state futures complete with an
     * {@link IllegalStateException}. This way, threads waiting for these
     * futures can continue. If the futures would not fail, threads waiting on
     * them would block indefinitely because after being destroyed, this channel
     * is not going to change its connection state any longer.
     * </p>
     * 
     * @since EPICS Jackie 2.0.0
     */
    protected void failWaitForConnectionStateFutures() {
        // This method is only supposed to be called after the channel has been
        // destroyed.
        if (!isDestroyed()) {
            throw new IllegalStateException(
                    "This method must be called after the channel has been put into the destroyed state.");
        }
        LinkedList<WaitForConnectionStateListenableFuture> localWaitForConnectedFutures;
        LinkedList<WaitForConnectionStateListenableFuture> localWaitForDisconnectedFutures;
        synchronized (connectionStateLock) {
            // In theory, just synchronizing on the connectionStateLock and then
            // using the list of futures should be fine. However, if this method
            // is called twice from some reason, this would lead to problems, so
            // we rather play it safe.
            localWaitForConnectedFutures = waitForConnectedFutures;
            waitForConnectedFutures = new LinkedList<WaitForConnectionStateListenableFuture>();
            localWaitForDisconnectedFutures = waitForDisconnectedFutures;
            waitForDisconnectedFutures = new LinkedList<WaitForConnectionStateListenableFuture>();
        }
        for (WaitForConnectionStateListenableFuture future : localWaitForConnectedFutures) {
            try {
                future.setExecutionException(new IllegalStateException(
                        "The channel has been destroyed while waiting for the connection state."));
            } catch (IllegalStateException e) {
                // When a future has just been added, the waitForConnectionState
                // method might already have failed the future. This is not a
                // problem, so we simply ignore this exception.
            }
        }
        for (WaitForConnectionStateListenableFuture future : localWaitForDisconnectedFutures) {
            try {
                future.setExecutionException(new IllegalStateException(
                        "The channel has been destroyed while waiting for the connection state."));
            } catch (IllegalStateException e) {
                // When a future has just been added, the waitForConnectionState
                // method might already have failed the future. This is not a
                // problem, so we simply ignore this exception.
            }
        }
    }

    /**
     * <p>
     * Updates the connection state and notifies the connection state listeners.
     * This method delegates calling each listener to
     * {@link #notifyConnectionListener(ChannelAccessConnectionListener, boolean)},
     * so that derived classes can take extra precautions (e.g. catch certain
     * exceptions thrown by a listener). Accordingly, setting the result of
     * futures created for calls to the {@link #waitForConnectionState(boolean)}
     * method is delegated to the
     * {@link #notifyWaitForConnectionStateFutureListener(ListenableFuture, FutureCompletionListener)}
     * method.
     * </p>
     * 
     * <p>
     * This method should be called by derived classes whenever the connection
     * state of the channel changes.
     * </p>
     * 
     * @param nowConnected
     *            <code>true</code> if the channel is now connected,
     *            <code>false</code> if it is now disconnected.
     */
    protected void updateConnectionState(boolean nowConnected) {
        ArrayList<ListenerRegistration> localListeners;
        LinkedList<WaitForConnectionStateListenableFuture> waitForConnectionStateFutures = null;
        synchronized (connectionStateLock) {
            // If the connection state has not changed, we do not want to notify
            // any listeners and return.
            if (connected == nowConnected) {
                return;
            }
            connected = nowConnected;
            if (nowConnected) {
                connectedOnce = true;
                if (!waitForConnectedFutures.isEmpty()) {
                    waitForConnectionStateFutures = waitForConnectedFutures;
                    waitForConnectedFutures = new LinkedList<WaitForConnectionStateListenableFuture>();
                }
            } else {
                if (!waitForDisconnectedFutures.isEmpty()) {
                    waitForConnectionStateFutures = waitForDisconnectedFutures;
                    waitForDisconnectedFutures = new LinkedList<WaitForConnectionStateListenableFuture>();
                }
            }
            // We make a copy of the current set of listeners. This way, we do
            // not have to hold the connectionListenerLock when calling the
            // listeners.
            localListeners = new ArrayList<ListenerRegistration>(
                    listeners.values());
        }
        // Now, that we have released the listenerLock, we can notify the
        // listenable futures and the connection listeners. The listenable
        // futures are only notified once, so we can do this without any lock,
        // and the connection listernes are protected by individual locks.
        if (waitForConnectionStateFutures != null) {
            for (SimpleListenableFuture<Boolean> future : waitForConnectionStateFutures) {
                future.setResult(Boolean.valueOf(nowConnected));
            }
        }
        for (ListenerRegistration listenerRegistration : localListeners) {
            // We have to synchronize on the per-listener lock while checking
            // the listener's registration and also keep the lock while
            // notifying the listener. This way, the listener cannot be removed
            // while we notify it and is never notified after it has been
            // removed.
            listenerRegistration.lock.lock();
            try {
                // We set the firstNotificationSet flag so that when the
                // listener has just been added, it does not receive an out of
                // order connection state. Without this flag, it could happen
                // that the listener is added, but before receiving its first
                // connection state (the current state at the time the listener
                // is added) a connection state change is triggered. In this
                // case, the older connection state would be sent to the
                // listener after the newer connection state.
                if (listenerRegistration.active) {
                    listenerRegistration.firstNotificationSent = true;
                    notifyConnectionListener(listenerRegistration.listener,
                            connected);
                }
            } finally {
                listenerRegistration.lock.unlock();
            }
        }
    }

    @Override
    public ListenerLockPolicy getListenerLockPolicy() {
        return listenerLockPolicy;
    }

    /**
     * Notifies a connection listener. This method can be overridden in order to
     * modify how the listener is called. For example, some implementations
     * might want to catch certain exceptions thrown by a listener.
     * 
     * @param listener
     *            the listener that should be notified. Notification should
     *            happen by calling the listeners
     *            {@link FutureCompletionListener#completed(ListenableFuture)
     *            completed()} method, passing a reference to this future
     *            instance.
     * @param nowConnected
     *            <code>true</code> if the channel is now connected,
     *            <code>false</code> if it is now disconnected.
     */
    protected void notifyConnectionListener(
            ChannelAccessConnectionListener listener, boolean nowConnected) {
        listener.connectionStateChanged(this, nowConnected);
    }

    /**
     * <p>
     * Notifies a completion listener that has been registered with a listenable
     * future returned by {@link #waitForConnectionState(boolean)}.
     * </p>
     * 
     * <p>
     * This method can be overridden in order to modify how the listener is
     * called. For example, some implementations might want to catch certain
     * exceptions thrown by a listener.
     * </p>
     * 
     * @param future
     *            future that was returned by
     *            {@link #waitForConnectionState(boolean)} that completed.
     * @param listener
     *            listener that has been registered with the <code>future</code>
     *            and shall be notify. This method must call the listener's
     *            <code>completed(...)</code> method, passing the
     *            <code>future</code>.
     */
    protected void notifyWaitForConnectionStateFutureListener(
            ListenableFuture<Boolean> future,
            FutureCompletionListener<? super Boolean> listener) {
        listener.completed(future);
    }

    @Override
    public boolean isConnected() {
        synchronized (connectionStateLock) {
            return connected;
        }
    }

    @Override
    public boolean isConnectedOnce() {
        synchronized (connectionStateLock) {
            return connectedOnce;
        }
    }

    @Override
    public boolean addConnectionListener(
            ChannelAccessConnectionListener listener) {
        if (listener == null) {
            throw new NullPointerException(
                    "The listener parameter must not be null.");
        }
        ListenerRegistration listenerRegistration;
        boolean localConnected;
        synchronized (connectionStateLock) {
            if (listeners.containsKey(listener)) {
                return false;
            }
            listenerRegistration = new ListenerRegistration(listener);
            listeners.put(listener, listenerRegistration);
            localConnected = connected;
        }
        // We notify the listener with the current connection state. This gives
        // the listener a chance to know the current connection state and work
        // with it until the connection state changes.
        // We do this after releasing the connectionStateLock. However, we
        // acquire the per-listener lock. This way, if a another thread removes
        // the listener while we are still in this method the listener will not
        // get removed before the notification has finished.
        // This also ensures that the listener will not receive the "current"
        // connection state, that we got while holding the connectionStateLock,
        // if it has already received an updated connection state (triggered by
        // updateConnectionState(...).
        // If we cannot acquire the lock, we simply continue. This is safe
        // because there are only two reasons why the lock might be held by a
        // different thread: First, the listener might be in the process of
        // being removed. Second, the listener might be in the process of being
        // notified with a new connection state. In either case, there is no
        // need for us to call the listener now.
        boolean gotLock = listenerRegistration.lock.tryLock();
        if (gotLock) {
            try {
                // The listener might have been removed before we entered the
                // synchronized block or it might have been notified with a
                // newer event, so we have to check both flags.
                if (listenerRegistration.active
                        && !listenerRegistration.firstNotificationSent) {
                    listenerRegistration.firstNotificationSent = true;
                    notifyConnectionListener(listener, localConnected);
                }
            } finally {
                listenerRegistration.lock.unlock();
            }
        }
        return true;
    }

    @Override
    public boolean removeConnectionListener(
            ChannelAccessConnectionListener listener) {
        ListenerRegistration listenerRegistration;
        synchronized (connectionStateLock) {
            listenerRegistration = listeners.remove(listener);
        }
        // We acquire the per-listener lock after releasing the listenerLock.
        // This way, we can avoid holding two locks at the same time.
        if (listenerRegistration != null) {
            boolean gotLock;
            if (listenerLockPolicy == ListenerLockPolicy.BLOCK) {
                listenerRegistration.lock.lock();
                gotLock = true;
            } else {
                gotLock = listenerRegistration.lock.tryLock();
            }
            if (gotLock) {
                try {
                    listenerRegistration.active = false;
                } finally {
                    listenerRegistration.lock.unlock();
                }
            } else {
                if (listenerLockPolicy == ListenerLockPolicy.REPORT) {
                    throw new ConcurrentNotificationException(
                            "The listener has been removed, but a notification is already in progress, so the listener is still going to be notified one more time.");
                }
            }
            return true;
        } else {
            return false;
        }
    }

    @Override
    public ListenableFuture<Boolean> waitForConnectionState(
            boolean connectionState) {
        WaitForConnectionStateListenableFuture future = new WaitForConnectionStateListenableFuture(
                this, listenerLockPolicy);
        // If this channel has been destroyed, we return a future with an
        // exception. This is the same effect as if the future had been added
        // just before the channel was destroyed.
        if (isDestroyed()) {
            future.setExecutionException(new IllegalStateException(
                    "The channel has been destroyed while waiting for the connection state."));
            return future;
        }
        synchronized (connectionStateLock) {
            // If we already are in the desired state, we can complete the
            // future right away. Otherwise, we will complete it when the
            // connection state changes.
            if (this.connected == connectionState) {
                future.setResult(Boolean.valueOf(connectionState));
            } else {
                if (connectionState) {
                    waitForConnectedFutures.add(future);
                } else {
                    waitForDisconnectedFutures.add(future);
                }
            }
        }
        // After we have added the future to the list, we check again whether
        // this channel has been destroyed. If it has been destroyed, we cancel
        // the future. We have to do this check because the channel might have
        // been destroyed between checking us the first time and adding the
        // future to the list.
        if (isDestroyed()) {
            try {
                future.setExecutionException(new IllegalStateException(
                        "The channel has been destroyed while waiting for the connection state."));
            } catch (IllegalStateException e) {
                // The future might already have completed. In this case we do
                // not need to set the exception and thus can simply continue.
            }
        }
        return future;
    }

    @Override
    public ListenableFuture<? extends ChannelAccessGettableValue<?>> get(
            ChannelAccessValueType valueType) {
        return get(valueType, 0);
    }

    @Override
    public ListenableFuture<? extends ChannelAccessGettableValue<?>> getNative() {
        return getNative(0);
    }

    @Override
    public ListenableFuture<? extends ChannelAccessGettableValue<?>> getNative(
            int count) {
        return get(null, count);
    }

    @Override
    public ListenableFuture<ChannelAccessSimpleOnlyString> getString() {
        return getString(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessSimpleOnlyString> getString(
            int count) {
        return (ListenableFuture<ChannelAccessSimpleOnlyString>) get(
                ChannelAccessValueType.DBR_STRING, count);
    }

    @Override
    public ListenableFuture<ChannelAccessAlarmOnlyString> getAlarmString() {
        return getAlarmString(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessAlarmOnlyString> getAlarmString(
            int count) {
        return (ListenableFuture<ChannelAccessAlarmOnlyString>) get(
                ChannelAccessValueType.DBR_STS_STRING, count);

    }

    @Override
    public ListenableFuture<ChannelAccessTimeString> getTimeString() {
        return getTimeString(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessTimeString> getTimeString(int count) {
        return (ListenableFuture<ChannelAccessTimeString>) get(
                ChannelAccessValueType.DBR_TIME_STRING, count);
    }

    @Override
    public ListenableFuture<ChannelAccessAlarmOnlyString> getGraphicsString() {
        return getGraphicsString(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessAlarmOnlyString> getGraphicsString(
            int count) {
        return (ListenableFuture<ChannelAccessAlarmOnlyString>) get(
                ChannelAccessValueType.DBR_GR_STRING, count);

    }

    @Override
    public ListenableFuture<ChannelAccessAlarmOnlyString> getControlsString() {
        return getControlsString(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessAlarmOnlyString> getControlsString(
            int count) {
        return (ListenableFuture<ChannelAccessAlarmOnlyString>) get(
                ChannelAccessValueType.DBR_CTRL_STRING, count);
    }

    @Override
    public ListenableFuture<ChannelAccessSimpleOnlyShort> getShort() {
        return getShort(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessSimpleOnlyShort> getShort(int count) {
        return (ListenableFuture<ChannelAccessSimpleOnlyShort>) get(
                ChannelAccessValueType.DBR_SHORT, count);
    }

    @Override
    public ListenableFuture<ChannelAccessAlarmOnlyShort> getAlarmShort() {
        return getAlarmShort(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessAlarmOnlyShort> getAlarmShort(
            int count) {
        return (ListenableFuture<ChannelAccessAlarmOnlyShort>) get(
                ChannelAccessValueType.DBR_STS_SHORT, count);
    }

    @Override
    public ListenableFuture<ChannelAccessTimeShort> getTimeShort() {
        return getTimeShort(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessTimeShort> getTimeShort(int count) {
        return (ListenableFuture<ChannelAccessTimeShort>) get(
                ChannelAccessValueType.DBR_TIME_SHORT, count);
    }

    @Override
    public ListenableFuture<ChannelAccessGraphicsOnlyShort> getGraphicsShort() {
        return getGraphicsShort(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessGraphicsOnlyShort> getGraphicsShort(
            int count) {
        return (ListenableFuture<ChannelAccessGraphicsOnlyShort>) get(
                ChannelAccessValueType.DBR_GR_SHORT, count);
    }

    @Override
    public ListenableFuture<ChannelAccessControlsShort> getControlsShort() {
        return getControlsShort(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessControlsShort> getControlsShort(
            int count) {
        return (ListenableFuture<ChannelAccessControlsShort>) get(
                ChannelAccessValueType.DBR_CTRL_SHORT, count);
    }

    @Override
    public ListenableFuture<ChannelAccessSimpleOnlyFloat> getFloat() {
        return getFloat(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessSimpleOnlyFloat> getFloat(int count) {
        return (ListenableFuture<ChannelAccessSimpleOnlyFloat>) get(
                ChannelAccessValueType.DBR_FLOAT, count);
    }

    @Override
    public ListenableFuture<ChannelAccessAlarmOnlyFloat> getAlarmFloat() {
        return getAlarmFloat(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessAlarmOnlyFloat> getAlarmFloat(
            int count) {
        return (ListenableFuture<ChannelAccessAlarmOnlyFloat>) get(
                ChannelAccessValueType.DBR_STS_FLOAT, count);
    }

    @Override
    public ListenableFuture<ChannelAccessTimeFloat> getTimeFloat() {
        return getTimeFloat(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessTimeFloat> getTimeFloat(int count) {
        return (ListenableFuture<ChannelAccessTimeFloat>) get(
                ChannelAccessValueType.DBR_TIME_FLOAT, count);
    }

    @Override
    public ListenableFuture<ChannelAccessGraphicsOnlyFloat> getGraphicsFloat() {
        return getGraphicsFloat(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessGraphicsOnlyFloat> getGraphicsFloat(
            int count) {
        return (ListenableFuture<ChannelAccessGraphicsOnlyFloat>) get(
                ChannelAccessValueType.DBR_GR_FLOAT, count);
    }

    @Override
    public ListenableFuture<ChannelAccessControlsFloat> getControlsFloat() {
        return getControlsFloat(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessControlsFloat> getControlsFloat(
            int count) {
        return (ListenableFuture<ChannelAccessControlsFloat>) get(
                ChannelAccessValueType.DBR_CTRL_FLOAT, count);
    }

    @Override
    public ListenableFuture<ChannelAccessSimpleOnlyEnum> getEnum() {
        return getEnum(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessSimpleOnlyEnum> getEnum(int count) {
        return (ListenableFuture<ChannelAccessSimpleOnlyEnum>) get(
                ChannelAccessValueType.DBR_ENUM, count);
    }

    @Override
    public ListenableFuture<ChannelAccessAlarmOnlyEnum> getAlarmEnum() {
        return getAlarmEnum(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessAlarmOnlyEnum> getAlarmEnum(
            int count) {
        return (ListenableFuture<ChannelAccessAlarmOnlyEnum>) get(
                ChannelAccessValueType.DBR_STS_ENUM, count);
    }

    @Override
    public ListenableFuture<ChannelAccessTimeEnum> getTimeEnum() {
        return getTimeEnum(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessTimeEnum> getTimeEnum(int count) {
        return (ListenableFuture<ChannelAccessTimeEnum>) get(
                ChannelAccessValueType.DBR_TIME_ENUM, count);
    }

    @Override
    public ListenableFuture<ChannelAccessGraphicsEnum> getGraphicsEnum() {
        return getGraphicsEnum(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessGraphicsEnum> getGraphicsEnum(
            int count) {
        return (ListenableFuture<ChannelAccessGraphicsEnum>) get(
                ChannelAccessValueType.DBR_GR_ENUM, count);
    }

    @Override
    public ListenableFuture<ChannelAccessGraphicsEnum> getControlsEnum() {
        return getControlsEnum(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessGraphicsEnum> getControlsEnum(
            int count) {
        return (ListenableFuture<ChannelAccessGraphicsEnum>) get(
                ChannelAccessValueType.DBR_CTRL_ENUM, count);
    }

    @Override
    public ListenableFuture<ChannelAccessSimpleOnlyChar> getChar() {
        return getChar(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessSimpleOnlyChar> getChar(int count) {
        return (ListenableFuture<ChannelAccessSimpleOnlyChar>) get(
                ChannelAccessValueType.DBR_CHAR, count);
    }

    @Override
    public ListenableFuture<ChannelAccessAlarmOnlyChar> getAlarmChar() {
        return getAlarmChar(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessAlarmOnlyChar> getAlarmChar(
            int count) {
        return (ListenableFuture<ChannelAccessAlarmOnlyChar>) get(
                ChannelAccessValueType.DBR_STS_CHAR, count);
    }

    @Override
    public ListenableFuture<ChannelAccessTimeChar> getTimeChar() {
        return getTimeChar(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessTimeChar> getTimeChar(int count) {
        return (ListenableFuture<ChannelAccessTimeChar>) get(
                ChannelAccessValueType.DBR_TIME_CHAR, count);
    }

    @Override
    public ListenableFuture<ChannelAccessGraphicsOnlyChar> getGraphicsChar() {
        return getGraphicsChar(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessGraphicsOnlyChar> getGraphicsChar(
            int count) {
        return (ListenableFuture<ChannelAccessGraphicsOnlyChar>) get(
                ChannelAccessValueType.DBR_GR_CHAR, count);
    }

    @Override
    public ListenableFuture<ChannelAccessControlsChar> getControlsChar() {
        return getControlsChar(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessControlsChar> getControlsChar(
            int count) {
        return (ListenableFuture<ChannelAccessControlsChar>) get(
                ChannelAccessValueType.DBR_CTRL_CHAR, count);
    }

    @Override
    public ListenableFuture<ChannelAccessSimpleOnlyLong> getLong() {
        return getLong(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessSimpleOnlyLong> getLong(int count) {
        return (ListenableFuture<ChannelAccessSimpleOnlyLong>) get(
                ChannelAccessValueType.DBR_LONG, count);
    }

    @Override
    public ListenableFuture<ChannelAccessAlarmOnlyLong> getAlarmLong() {
        return getAlarmLong(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessAlarmOnlyLong> getAlarmLong(
            int count) {
        return (ListenableFuture<ChannelAccessAlarmOnlyLong>) get(
                ChannelAccessValueType.DBR_STS_LONG, count);
    }

    @Override
    public ListenableFuture<ChannelAccessTimeLong> getTimeLong() {
        return getTimeLong(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessTimeLong> getTimeLong(int count) {
        return (ListenableFuture<ChannelAccessTimeLong>) get(
                ChannelAccessValueType.DBR_TIME_LONG, count);
    }

    @Override
    public ListenableFuture<ChannelAccessGraphicsOnlyLong> getGraphicsLong() {
        return getGraphicsLong(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessGraphicsOnlyLong> getGraphicsLong(
            int count) {
        return (ListenableFuture<ChannelAccessGraphicsOnlyLong>) get(
                ChannelAccessValueType.DBR_GR_LONG, count);
    }

    @Override
    public ListenableFuture<ChannelAccessControlsLong> getControlsLong() {
        return getControlsLong(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessControlsLong> getControlsLong(
            int count) {
        return (ListenableFuture<ChannelAccessControlsLong>) get(
                ChannelAccessValueType.DBR_CTRL_LONG, count);
    }

    @Override
    public ListenableFuture<ChannelAccessSimpleOnlyDouble> getDouble() {
        return getDouble(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessSimpleOnlyDouble> getDouble(
            int count) {
        return (ListenableFuture<ChannelAccessSimpleOnlyDouble>) get(
                ChannelAccessValueType.DBR_DOUBLE, count);
    }

    @Override
    public ListenableFuture<ChannelAccessAlarmOnlyDouble> getAlarmDouble() {
        return getAlarmDouble(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessAlarmOnlyDouble> getAlarmDouble(
            int count) {
        return (ListenableFuture<ChannelAccessAlarmOnlyDouble>) get(
                ChannelAccessValueType.DBR_STS_DOUBLE, count);
    }

    @Override
    public ListenableFuture<ChannelAccessTimeDouble> getTimeDouble() {
        return getTimeDouble(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessTimeDouble> getTimeDouble(int count) {
        return (ListenableFuture<ChannelAccessTimeDouble>) get(
                ChannelAccessValueType.DBR_TIME_DOUBLE, count);
    }

    @Override
    public ListenableFuture<ChannelAccessGraphicsOnlyDouble> getGraphicsDouble() {
        return getGraphicsDouble(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessGraphicsOnlyDouble> getGraphicsDouble(
            int count) {
        return (ListenableFuture<ChannelAccessGraphicsOnlyDouble>) get(
                ChannelAccessValueType.DBR_GR_DOUBLE, count);
    }

    @Override
    public ListenableFuture<ChannelAccessControlsDouble> getControlsDouble() {
        return getControlsDouble(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessControlsDouble> getControlsDouble(
            int count) {
        return (ListenableFuture<ChannelAccessControlsDouble>) get(
                ChannelAccessValueType.DBR_CTRL_DOUBLE, count);
    }

    @Override
    public ListenableFuture<ChannelAccessAlarmAcknowledgementStatus> getAlarmAcknowledgementStatus() {
        return getAlarmAcknowledgementStatus(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessAlarmAcknowledgementStatus> getAlarmAcknowledgementStatus(
            int count) {
        return (ListenableFuture<ChannelAccessAlarmAcknowledgementStatus>) get(
                ChannelAccessValueType.DBR_STSACK_STRING, count);
    }

    @Override
    public ListenableFuture<ChannelAccessClassName> getClassName() {
        return getClassName(0);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ListenableFuture<ChannelAccessClassName> getClassName(int count) {
        return (ListenableFuture<ChannelAccessClassName>) get(
                ChannelAccessValueType.DBR_CLASS_NAME, count);
    }

    @Override
    public ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitor(
            ChannelAccessValueType valueType) {
        return monitor(valueType, 0);
    }

    @Override
    public ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitor(
            ChannelAccessValueType valueType,
            ChannelAccessEventMask eventMask) {
        return monitor(valueType, 0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitor(
            ChannelAccessValueType valueType, int count) {
        return monitor(valueType, count, null);
    }

    @Override
    public ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitorNative() {
        return monitorNative(0);
    }

    @Override
    public ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitorNative(
            ChannelAccessEventMask eventMask) {
        return monitorNative(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitorNative(
            int count) {
        return monitorNative(count, null);
    }

    @Override
    public ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitorNative(
            int count, ChannelAccessEventMask eventMask) {
        return monitor(null, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyString> monitorString() {
        return monitorString(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyString> monitorString(
            ChannelAccessEventMask eventMask) {
        return monitorString(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyString> monitorString(
            int count) {
        return monitorString(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyString> monitorString(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessSimpleOnlyString>) monitor(
                ChannelAccessValueType.DBR_STRING, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorAlarmString() {
        return monitorAlarmString(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorAlarmString(
            ChannelAccessEventMask eventMask) {
        return monitorAlarmString(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorAlarmString(
            int count) {
        return monitorAlarmString(0, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorAlarmString(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessAlarmOnlyString>) monitor(
                ChannelAccessValueType.DBR_STS_STRING, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessTimeString> monitorTimeString() {
        return monitorTimeString(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessTimeString> monitorTimeString(
            ChannelAccessEventMask eventMask) {
        return monitorTimeString(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessTimeString> monitorTimeString(
            int count) {
        return monitorTimeString(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessTimeString> monitorTimeString(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessTimeString>) monitor(
                ChannelAccessValueType.DBR_TIME_STRING, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorGraphicsString() {
        return monitorGraphicsString(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorGraphicsString(
            ChannelAccessEventMask eventMask) {
        return monitorGraphicsString(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorGraphicsString(
            int count) {
        return monitorGraphicsString(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorGraphicsString(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessAlarmOnlyString>) monitor(
                ChannelAccessValueType.DBR_GR_STRING, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorControlsString() {
        return monitorControlsString(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorControlsString(
            ChannelAccessEventMask eventMask) {
        return monitorControlsString(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorControlsString(
            int count) {
        return monitorControlsString(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorControlsString(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessAlarmOnlyString>) monitor(
                ChannelAccessValueType.DBR_CTRL_STRING, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyShort> monitorShort() {
        return monitorShort(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyShort> monitorShort(
            ChannelAccessEventMask eventMask) {
        return monitorShort(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyShort> monitorShort(
            int count) {
        return monitorShort(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyShort> monitorShort(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessSimpleOnlyShort>) monitor(
                ChannelAccessValueType.DBR_SHORT, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyShort> monitorAlarmShort() {
        return monitorAlarmShort(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyShort> monitorAlarmShort(
            ChannelAccessEventMask eventMask) {
        return monitorAlarmShort(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyShort> monitorAlarmShort(
            int count) {
        return monitorAlarmShort(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyShort> monitorAlarmShort(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessAlarmOnlyShort>) monitor(
                ChannelAccessValueType.DBR_STS_SHORT, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessTimeShort> monitorTimeShort() {
        return monitorTimeShort(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessTimeShort> monitorTimeShort(
            ChannelAccessEventMask eventMask) {
        return monitorTimeShort(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessTimeShort> monitorTimeShort(
            int count) {
        return monitorTimeShort(0, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessTimeShort> monitorTimeShort(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessTimeShort>) monitor(
                ChannelAccessValueType.DBR_TIME_SHORT, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsOnlyShort> monitorGraphicsShort() {
        return monitorGraphicsShort(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsOnlyShort> monitorGraphicsShort(
            ChannelAccessEventMask eventMask) {
        return monitorGraphicsShort(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsOnlyShort> monitorGraphicsShort(
            int count) {
        return monitorGraphicsShort(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsOnlyShort> monitorGraphicsShort(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessGraphicsOnlyShort>) monitor(
                ChannelAccessValueType.DBR_GR_SHORT, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessControlsShort> monitorControlsShort() {
        return monitorControlsShort(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessControlsShort> monitorControlsShort(
            ChannelAccessEventMask eventMask) {
        return monitorControlsShort(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessControlsShort> monitorControlsShort(
            int count) {
        return monitorControlsShort(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessControlsShort> monitorControlsShort(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessControlsShort>) monitor(
                ChannelAccessValueType.DBR_CTRL_SHORT, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyFloat> monitorFloat() {
        return monitorFloat(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyFloat> monitorFloat(
            ChannelAccessEventMask eventMask) {
        return monitorFloat(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyFloat> monitorFloat(
            int count) {
        return monitorFloat(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyFloat> monitorFloat(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessSimpleOnlyFloat>) monitor(
                ChannelAccessValueType.DBR_FLOAT, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyFloat> monitorAlarmFloat() {
        return monitorAlarmFloat(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyFloat> monitorAlarmFloat(
            ChannelAccessEventMask eventMask) {
        return monitorAlarmFloat(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyFloat> monitorAlarmFloat(
            int count) {
        return monitorAlarmFloat(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyFloat> monitorAlarmFloat(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessAlarmOnlyFloat>) monitor(
                ChannelAccessValueType.DBR_STS_FLOAT, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessTimeFloat> monitorTimeFloat() {
        return monitorTimeFloat(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessTimeFloat> monitorTimeFloat(
            ChannelAccessEventMask eventMask) {
        return monitorTimeFloat(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessTimeFloat> monitorTimeFloat(
            int count) {
        return monitorTimeFloat(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessTimeFloat> monitorTimeFloat(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessTimeFloat>) monitor(
                ChannelAccessValueType.DBR_TIME_FLOAT, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsOnlyFloat> monitorGraphicsFloat() {
        return monitorGraphicsFloat(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsOnlyFloat> monitorGraphicsFloat(
            ChannelAccessEventMask eventMask) {
        return monitorGraphicsFloat(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsOnlyFloat> monitorGraphicsFloat(
            int count) {
        return monitorGraphicsFloat(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsOnlyFloat> monitorGraphicsFloat(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessGraphicsOnlyFloat>) monitor(
                ChannelAccessValueType.DBR_GR_FLOAT, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessControlsFloat> monitorControlsFloat() {
        return monitorControlsFloat(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessControlsFloat> monitorControlsFloat(
            ChannelAccessEventMask eventMask) {
        return monitorControlsFloat(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessControlsFloat> monitorControlsFloat(
            int count) {
        return monitorControlsFloat(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessControlsFloat> monitorControlsFloat(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessControlsFloat>) monitor(
                ChannelAccessValueType.DBR_CTRL_FLOAT, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyEnum> monitorEnum() {
        return monitorEnum(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyEnum> monitorEnum(
            ChannelAccessEventMask eventMask) {
        return monitorEnum(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyEnum> monitorEnum(
            int count) {
        return monitorEnum(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyEnum> monitorEnum(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessSimpleOnlyEnum>) monitor(
                ChannelAccessValueType.DBR_ENUM, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyEnum> monitorAlarmEnum() {
        return monitorAlarmEnum(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyEnum> monitorAlarmEnum(
            ChannelAccessEventMask eventMask) {
        return monitorAlarmEnum(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyEnum> monitorAlarmEnum(
            int count) {
        return monitorAlarmEnum(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyEnum> monitorAlarmEnum(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessAlarmOnlyEnum>) monitor(
                ChannelAccessValueType.DBR_STS_ENUM, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessTimeEnum> monitorTimeEnum() {
        return monitorTimeEnum(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessTimeEnum> monitorTimeEnum(
            ChannelAccessEventMask eventMask) {
        return monitorTimeEnum(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessTimeEnum> monitorTimeEnum(
            int count) {
        return monitorTimeEnum(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessTimeEnum> monitorTimeEnum(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessTimeEnum>) monitor(
                ChannelAccessValueType.DBR_TIME_ENUM, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorGraphicsEnum() {
        return monitorGraphicsEnum(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorGraphicsEnum(
            ChannelAccessEventMask eventMask) {
        return monitorGraphicsEnum(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorGraphicsEnum(
            int count) {
        return monitorGraphicsEnum(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorGraphicsEnum(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessGraphicsEnum>) monitor(
                ChannelAccessValueType.DBR_GR_ENUM, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorControlsEnum() {
        return monitorControlsEnum(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorControlsEnum(
            ChannelAccessEventMask eventMask) {
        return monitorControlsEnum(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorControlsEnum(
            int count) {
        return monitorControlsEnum(0, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorControlsEnum(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessGraphicsEnum>) monitor(
                ChannelAccessValueType.DBR_CTRL_ENUM, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyChar> monitorChar() {
        return monitorChar(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyChar> monitorChar(
            ChannelAccessEventMask eventMask) {
        return monitorChar(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyChar> monitorChar(
            int count) {
        return monitorChar(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyChar> monitorChar(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessSimpleOnlyChar>) monitor(
                ChannelAccessValueType.DBR_CHAR, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyChar> monitorAlarmChar() {
        return monitorAlarmChar(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyChar> monitorAlarmChar(
            ChannelAccessEventMask eventMask) {
        return monitorAlarmChar(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyChar> monitorAlarmChar(
            int count) {
        return monitorAlarmChar(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyChar> monitorAlarmChar(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessAlarmOnlyChar>) monitor(
                ChannelAccessValueType.DBR_STS_CHAR, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessTimeChar> monitorTimeChar() {
        return monitorTimeChar(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessTimeChar> monitorTimeChar(
            ChannelAccessEventMask eventMask) {
        return monitorTimeChar(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessTimeChar> monitorTimeChar(
            int count) {
        return monitorTimeChar(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessTimeChar> monitorTimeChar(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessTimeChar>) monitor(
                ChannelAccessValueType.DBR_TIME_CHAR, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsOnlyChar> monitorGraphicsChar() {
        return monitorGraphicsChar(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsOnlyChar> monitorGraphicsChar(
            ChannelAccessEventMask eventMask) {
        return monitorGraphicsChar(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsOnlyChar> monitorGraphicsChar(
            int count) {
        return monitorGraphicsChar(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsOnlyChar> monitorGraphicsChar(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessGraphicsOnlyChar>) monitor(
                ChannelAccessValueType.DBR_GR_CHAR, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessControlsChar> monitorControlsChar() {
        return monitorControlsChar(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessControlsChar> monitorControlsChar(
            ChannelAccessEventMask eventMask) {
        return monitorControlsChar(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessControlsChar> monitorControlsChar(
            int count) {
        return monitorControlsChar(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessControlsChar> monitorControlsChar(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessControlsChar>) monitor(
                ChannelAccessValueType.DBR_CTRL_CHAR, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyLong> monitorLong() {
        return monitorLong(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyLong> monitorLong(
            ChannelAccessEventMask eventMask) {
        return monitorLong(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyLong> monitorLong(
            int count) {
        return monitorLong(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyLong> monitorLong(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessSimpleOnlyLong>) monitor(
                ChannelAccessValueType.DBR_LONG, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyLong> monitorAlarmLong() {
        return monitorAlarmLong(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyLong> monitorAlarmLong(
            ChannelAccessEventMask eventMask) {
        return monitorAlarmLong(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyLong> monitorAlarmLong(
            int count) {
        return monitorAlarmLong(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyLong> monitorAlarmLong(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessAlarmOnlyLong>) monitor(
                ChannelAccessValueType.DBR_STS_LONG, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessTimeLong> monitorTimeLong() {
        return monitorTimeLong(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessTimeLong> monitorTimeLong(
            ChannelAccessEventMask eventMask) {
        return monitorTimeLong(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessTimeLong> monitorTimeLong(
            int count) {
        return monitorTimeLong(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessTimeLong> monitorTimeLong(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessTimeLong>) monitor(
                ChannelAccessValueType.DBR_TIME_LONG, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsOnlyLong> monitorGraphicsLong() {
        return monitorGraphicsLong(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsOnlyLong> monitorGraphicsLong(
            ChannelAccessEventMask eventMask) {
        return monitorGraphicsLong(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsOnlyLong> monitorGraphicsLong(
            int count) {
        return monitorGraphicsLong(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsOnlyLong> monitorGraphicsLong(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessGraphicsOnlyLong>) monitor(
                ChannelAccessValueType.DBR_GR_LONG, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessControlsLong> monitorControlsLong() {
        return monitorControlsLong(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessControlsLong> monitorControlsLong(
            ChannelAccessEventMask eventMask) {
        return monitorControlsLong(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessControlsLong> monitorControlsLong(
            int count) {
        return monitorControlsLong(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessControlsLong> monitorControlsLong(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessControlsLong>) monitor(
                ChannelAccessValueType.DBR_CTRL_LONG, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyDouble> monitorDouble() {
        return monitorDouble(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyDouble> monitorDouble(
            ChannelAccessEventMask eventMask) {
        return monitorDouble(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyDouble> monitorDouble(
            int count) {
        return monitorDouble(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessSimpleOnlyDouble> monitorDouble(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessSimpleOnlyDouble>) monitor(
                ChannelAccessValueType.DBR_DOUBLE, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyDouble> monitorAlarmDouble() {
        return monitorAlarmDouble(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyDouble> monitorAlarmDouble(
            ChannelAccessEventMask eventMask) {
        return monitorAlarmDouble(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyDouble> monitorAlarmDouble(
            int count) {
        return monitorAlarmDouble(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmOnlyDouble> monitorAlarmDouble(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessAlarmOnlyDouble>) monitor(
                ChannelAccessValueType.DBR_STS_DOUBLE, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessTimeDouble> monitorTimeDouble() {
        return monitorTimeDouble(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessTimeDouble> monitorTimeDouble(
            ChannelAccessEventMask eventMask) {
        return monitorTimeDouble(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessTimeDouble> monitorTimeDouble(
            int count) {
        return monitorTimeDouble(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessTimeDouble> monitorTimeDouble(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessTimeDouble>) monitor(
                ChannelAccessValueType.DBR_TIME_DOUBLE, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsOnlyDouble> monitorGraphicsDouble() {
        return monitorGraphicsDouble(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsOnlyDouble> monitorGraphicsDouble(
            ChannelAccessEventMask eventMask) {
        return monitorGraphicsDouble(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsOnlyDouble> monitorGraphicsDouble(
            int count) {
        return monitorGraphicsDouble(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessGraphicsOnlyDouble> monitorGraphicsDouble(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessGraphicsOnlyDouble>) monitor(
                ChannelAccessValueType.DBR_GR_DOUBLE, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessControlsDouble> monitorControlsDouble() {
        return monitorControlsDouble(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessControlsDouble> monitorControlsDouble(
            ChannelAccessEventMask eventMask) {
        return monitorControlsDouble(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessControlsDouble> monitorControlsDouble(
            int count) {
        return monitorControlsDouble(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessControlsDouble> monitorControlsDouble(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessControlsDouble>) monitor(
                ChannelAccessValueType.DBR_CTRL_DOUBLE, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmAcknowledgementStatus> monitorAlarmAcknowledgementStatus() {
        return monitorAlarmAcknowledgementStatus(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmAcknowledgementStatus> monitorAlarmAcknowledgementStatus(
            ChannelAccessEventMask eventMask) {
        return monitorAlarmAcknowledgementStatus(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmAcknowledgementStatus> monitorAlarmAcknowledgementStatus(
            int count) {
        return monitorAlarmAcknowledgementStatus(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessAlarmAcknowledgementStatus> monitorAlarmAcknowledgementStatus(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessAlarmAcknowledgementStatus>) monitor(
                ChannelAccessValueType.DBR_STSACK_STRING, count, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessClassName> monitorClassName() {
        return monitorClassName(0);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessClassName> monitorClassName(
            ChannelAccessEventMask eventMask) {
        return monitorClassName(0, eventMask);
    }

    @Override
    public ChannelAccessMonitor<ChannelAccessClassName> monitorClassName(
            int count) {
        return monitorClassName(count, null);
    }

    @SuppressWarnings("unchecked")
    @Override
    public ChannelAccessMonitor<ChannelAccessClassName> monitorClassName(
            int count, ChannelAccessEventMask eventMask) {
        return (ChannelAccessMonitor<ChannelAccessClassName>) monitor(
                ChannelAccessValueType.DBR_CLASS_NAME, count, eventMask);
    }

    @Override
    public ListenableFuture<Void> putString(String value) {
        return putString(Collections.singleton(value));
    }

    @Override
    public ListenableFuture<Void> putString(String[] value) {
        return putString(Arrays.asList(value));
    }

    @Override
    public ListenableFuture<Void> putString(Collection<String> value) {
        return put(ChannelAccessValueFactory.createString(value,
                getClient().getConfiguration().getCharset()));
    }

    @Override
    public ListenableFuture<Void> putShort(short value) {
        return putShort(new short[] { value });
    }

    @Override
    public ListenableFuture<Void> putShort(short[] value) {
        return put(ChannelAccessValueFactory.createShort(value));
    }

    @Override
    public ListenableFuture<Void> putFloat(float value) {
        return putFloat(new float[] { value });
    }

    @Override
    public ListenableFuture<Void> putFloat(float[] value) {
        return put(ChannelAccessValueFactory.createFloat(value));
    }

    @Override
    public ListenableFuture<Void> putEnum(short value) {
        return putEnum(new short[] { value });
    }

    @Override
    public ListenableFuture<Void> putEnum(short[] value) {
        return put(ChannelAccessValueFactory.createEnum(value));
    }

    @Override
    public ListenableFuture<Void> putChar(byte value) {
        return putChar(new byte[] { value });
    }

    @Override
    public ListenableFuture<Void> putChar(byte[] value) {
        return put(ChannelAccessValueFactory.createChar(value));
    }

    @Override
    public ListenableFuture<Void> putLong(int value) {
        return putLong(new int[] { value });
    }

    @Override
    public ListenableFuture<Void> putLong(int[] value) {
        return put(ChannelAccessValueFactory.createLong(value));
    }

    @Override
    public ListenableFuture<Void> putDouble(double value) {
        return putDouble(new double[] { value });
    }

    @Override
    public ListenableFuture<Void> putDouble(double[] value) {
        return put(ChannelAccessValueFactory.createDouble(value));
    }

    @Override
    public ListenableFuture<Void> putConfigureAcknowledgeTransientAlarms(
            boolean explicitlyAcknowledgeTransientAlarms) {
        return put(ChannelAccessValueFactory
                .createConfigureAcknowledgeTransientAlarms(
                        explicitlyAcknowledgeTransientAlarms));
    }

    @Override
    public ListenableFuture<Void> putAcknowledgeAlarm(
            ChannelAccessAlarmSeverity acknowledgeAlarmSeverity) {
        return put(ChannelAccessValueFactory
                .createAcknowledgeAlarm(acknowledgeAlarmSeverity));
    }

}
