/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client;

import java.net.InetSocketAddress;
import java.util.LinkedList;
import java.util.Queue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;

import org.apache.commons.lang3.StringEscapeUtils;
import org.apache.commons.lang3.tuple.Pair;

import com.aquenos.epics.jackie.client.internal.CommunicationThread;
import com.aquenos.epics.jackie.client.internal.SingleProcessorCommunicationThread;
import com.aquenos.epics.jackie.common.exception.ErrorHandler;
import com.aquenos.epics.jackie.common.io.CommunicationProcessor;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus;
import com.aquenos.epics.jackie.common.util.FutureCompletionListener;
import com.aquenos.epics.jackie.common.util.IdentityWrapper;
import com.aquenos.epics.jackie.common.util.ListenableFuture;
import com.aquenos.epics.jackie.common.value.ChannelAccessGettableValue;

/**
 * Simple threading strategy that uses one thread for channel name resolution
 * and processing beacon messages and creates on additional thread for each
 * connection with a Channel Access server. By default, this strategy processes
 * listeners (e.g. for event monitors) within the communication thread that
 * triggers them. However, it can also be configured to use an
 * {@link ExecutorService} for delivering notifications.
 * 
 * @author Sebastian Marsching
 */
public class DefaultClientThreadingStrategy implements ClientThreadingStrategy {

    /**
     * Queue of notifications. This class is used internally for remembering
     * which notifications have been submitted for a certain event-source /
     * listener combination, so that the notifications are delivered in the
     * right order.
     * 
     * @author Sebastian Marsching
     */
    private static class NotificationQueue {

        private boolean taskIsRunning;
        private Queue<Runnable> tasks = new LinkedList<Runnable>();
        private int executionAttemptsWhileRunning;

    }

    private volatile boolean destroyed = false;
    private ErrorHandler errorHandler;
    private CommunicationThread beaconAndResolverThread;
    private ExecutorService notificationExecutorService;
    private ConcurrentHashMap<Object, NotificationQueue> notificationQueues;
    private boolean shutdownNotificationExecutorServiceOnDestruction;
    private ConcurrentHashMap<IdentityWrapper<CommunicationProcessor>, CommunicationThread> serverThreads = new ConcurrentHashMap<IdentityWrapper<CommunicationProcessor>, CommunicationThread>();

    /**
     * <p>
     * Creates a default threading strategy that uses the specified error
     * handler.
     * </p>
     * 
     * <p>
     * The error handler is used when one of the threads that is managed by this
     * strategy encounters an uncaught exception.
     * </p>
     * 
     * <p>
     * Using this constructor is equivalent to using
     * {@link #DefaultClientThreadingStrategy(ErrorHandler, ExecutorService, boolean)
     * DefaultClientThreadingStrategy(errorHandler, null, false)}.
     * </p>
     * 
     * @param errorHandler
     *            error handler to be used by this threading strategy. This
     *            error handler is used if an exception is thrown by one of the
     *            communication processors that are registered with this
     *            strategy. Typically, processors should handle exceptions
     *            themselves, so this is only a fall-back solution.
     */
    public DefaultClientThreadingStrategy(ErrorHandler errorHandler) {
        this(errorHandler, null, false);
    }

    /**
     * <p>
     * Creates a default threading strategy that uses the specified error
     * handler and executor service.
     * </p>
     * 
     * <p>
     * The error handler is used when one of the threads that is managed by this
     * strategy encounters an uncaught exception.
     * </p>
     * 
     * <p>
     * If the supplied <code>notificationExecutorService</code> is not
     * <code>null</code>, it is used when one of this strategy's listener
     * notification methods is called. Instead of calling the corresponding
     * listener method directly, a task is scheduled with the executor service
     * in order to call the listener method in a separate thread at a later
     * time. This class ensures that notifications are still delivered in the
     * right order, regardless of how the executor service schedules the tasks
     * submitted to it. If the <code>notificationExecutorService</code> is
     * <code>null</code>, the listener's methods are called right in the thread
     * calling this strategy's notification method.
     * </p>
     * 
     * <p>
     * For performance reasons, a <code>notificationExecutorService</code>
     * should only be used if it is really necessary. In general, notifying
     * listeners right in the calling thread is the best choice. Using separate
     * threads should only be considered if listeners might act erratically (by
     * blocking or running for an extended amount of time) and fixing the
     * listeners is not feasible. Due to internal optimizations, setting the
     * <code>notificationExecutorService</code> to <code>null</code> is much
     * more efficient than providing an executor service that runs its submitted
     * tasks in the calling thread.
     * </p>
     * 
     * <p>
     * If <code>shutdownNotificationExecutorServiceOnDestruction</code> is
     * <code>true</code> and <code>notificationExecutorService</code> is not
     * <code>null</code>, the <code>notificationExecutorService</code>'s
     * <code>shutdown()</code> method is called when this strategy's
     * <code>destroy()</code> method is called. Otherwise, the executor service
     * is not shutdown when this threading strategy is destroyed. The latter
     * behavior can be useful if the executor service that is passed to this
     * constructor is shared with another component that should stay operative
     * when this threading strategy is destroyed.
     * </p>
     * 
     * @param errorHandler
     *            error handler to be used by this threading strategy. This
     *            error handler is used if an exception is thrown by one of the
     *            communication processors that are registered with this
     *            strategy. Typically, processors should handle exceptions
     *            themselves, so this is only a fall-back solution. The error
     *            handler is also used when the
     *            <code>notificationExecutorService</code> is not
     *            <code>null</code> and one of the listener methods invoked
     *            through this execution service throws an exception.
     * @param notificationExecutorService
     *            executor service used to run listener notifications. If
     *            <code>null</code>, listeners are notified right in the calling
     *            thread.
     * @param shutdownNotificationExecutorServiceOnDestruction
     *            if <code>true</code>, the
     *            <code>notificationExecutorService</code> is shutdown when this
     *            strategy is destroyed. If <code>false</code>, the
     *            <code>notificationExecutorService</code> is kept alive past
     *            the destruction of this strategy.
     */
    public DefaultClientThreadingStrategy(ErrorHandler errorHandler,
            ExecutorService notificationExecutorService,
            boolean shutdownNotificationExecutorServiceOnDestruction) {
        if (errorHandler == null) {
            throw new NullPointerException();
        }
        this.errorHandler = errorHandler;
        this.notificationExecutorService = notificationExecutorService;
        this.shutdownNotificationExecutorServiceOnDestruction = shutdownNotificationExecutorServiceOnDestruction;
        if (this.notificationExecutorService != null) {
            this.notificationQueues = new ConcurrentHashMap<Object, NotificationQueue>();
        }
        this.beaconAndResolverThread = new CommunicationThread(
                "beacon-detector-and-channel-name-resolver", errorHandler);
        // We can start the thread right now. As long as no channel name
        // resolver is registered, it will simply sleep.
        this.beaconAndResolverThread.start();
    }

    @Override
    public void addServerConnection(InetSocketAddress serverAddress,
            CommunicationProcessor communicationProcessor) {
        if (destroyed) {
            throw new IllegalStateException(
                    "Cannot add a server after the threading strategy has been destroyed.");
        }
        StringBuilder sb = new StringBuilder();
        sb.append("ca-server-");
        sb.append(serverAddress.getAddress().getHostAddress());
        sb.append(':');
        sb.append(serverAddress.getPort());
        CommunicationThread thread = new SingleProcessorCommunicationThread(
                sb.toString(), communicationProcessor, errorHandler);
        if (serverThreads.putIfAbsent(
                IdentityWrapper.wrap(communicationProcessor), thread) != null) {
            // The specified communication processor has already been added
            // earlier. Therefore, there is no need to create another thread.
            // We simply do not start the thread so that the thread object we
            // just created gets garbage collected.
            return;
        }
        // If the threading strategy is destroyed after adding the thread, we do
        // not start the thread. This ensures that all threads that have been
        // started are stopped when destroying the threading strategy.
        if (!destroyed) {
            thread.start();
        }
    }

    @Override
    public void removeServerConnection(InetSocketAddress serverAddress,
            CommunicationProcessor communicationProcessor) {
        CommunicationThread thread = serverThreads.remove(IdentityWrapper
                .wrap(communicationProcessor));
        if (thread != null) {
            thread.stopGracefully();
        }
    }

    @Override
    public void notifyConnectionListener(final ChannelAccessChannel channel,
            final ChannelAccessConnectionListener listener,
            final boolean nowConnected) {
        if (notificationExecutorService == null) {
            listener.connectionStateChanged(channel, nowConnected);
        } else {
            Runnable notificationTask = new Runnable() {
                @Override
                public void run() {
                    // If the listener throws, we want to handle this right
                    // here. We do not know which executor service
                    // implementation is used and thus cannot know how it would
                    // handle an exception.
                    try {
                        listener.connectionStateChanged(channel, nowConnected);
                    } catch (Throwable t) {
                        errorHandler.handleError(
                                this.getClass(),
                                t,
                                "Connection listener for channel \""
                                        + StringEscapeUtils.escapeJava(channel
                                                .getName())
                                        + "\" threw an exception: "
                                        + t.getMessage());
                    }
                }
            };
            // We use the combined identities of the channel and the listener as
            // the synchronization key. This ensures that notifications from the
            // same channel for the same listener are delivered in order. We
            // want to compare by identity because we cannot possibly know how a
            // channel or a listener implementation might implement equals(...)
            // and hashCode().
            submitNotificationTask(
                    Pair.of(IdentityWrapper.wrap(channel),
                            IdentityWrapper.wrap(listener)), notificationTask);
        }
    }

    @Override
    public <T extends ChannelAccessGettableValue<?>> void notifyMonitorListenerError(
            final ChannelAccessMonitor<? extends T> monitor,
            final ChannelAccessMonitorListener<? super T> listener,
            final ChannelAccessStatus status, final String message) {
        if (notificationExecutorService == null) {
            listener.monitorError(monitor, status, message);
        } else {
            Runnable notificationTask = new Runnable() {
                @Override
                public void run() {
                    // If the listener throws, we want to handle this right
                    // here. We do not know which executor service
                    // implementation is used and thus cannot know how it would
                    // handle an exception.
                    try {
                        listener.monitorError(monitor, status, message);
                    } catch (Throwable t) {
                        errorHandler.handleError(
                                this.getClass(),
                                t,
                                "Monitor listener for channel \""
                                        + StringEscapeUtils.escapeJava(monitor
                                                .getChannel().getName())
                                        + "\" threw an exception: "
                                        + t.getMessage());
                    }
                }
            };
            // We use the combined identities of the monitor and the listener as
            // the synchronization key. This ensures that notifications from the
            // same monitor for the same listener are delivered in order. We
            // want to compare by identity because we cannot possibly know how a
            // monitor or a listener implementation might implement equals(...)
            // and hashCode().
            submitNotificationTask(
                    Pair.of(IdentityWrapper.wrap(monitor),
                            IdentityWrapper.wrap(listener)), notificationTask);
        }
    }

    @Override
    public <T extends ChannelAccessGettableValue<?>> void notifyMonitorListenerEvent(
            final ChannelAccessMonitor<? extends T> monitor,
            final ChannelAccessMonitorListener<? super T> listener,
            final T value) {
        if (notificationExecutorService == null) {
            listener.monitorEvent(monitor, value);
        } else {
            Runnable notificationTask = new Runnable() {
                @Override
                public void run() {
                    // If the listener throws, we want to handle this right
                    // here. We do not know which executor service
                    // implementation is used and thus cannot know how it would
                    // handle an exception.
                    try {
                        listener.monitorEvent(monitor, value);
                    } catch (Throwable t) {
                        errorHandler.handleError(
                                this.getClass(),
                                t,
                                "Monitor listener for channel \""
                                        + StringEscapeUtils.escapeJava(monitor
                                                .getChannel().getName())
                                        + "\" threw an exception: "
                                        + t.getMessage());
                    }
                }
            };
            // We use the combined identities of the monitor and the listener as
            // the synchronization key. This ensures that notifications from the
            // same monitor for the same listener are delivered in order. We
            // want to compare by identity because we cannot possibly know how a
            // monitor or a listener implementation might implement equals(...)
            // and hashCode().
            submitNotificationTask(
                    Pair.of(IdentityWrapper.wrap(monitor),
                            IdentityWrapper.wrap(listener)), notificationTask);
        }
    }

    @Override
    public <T> void notifyFutureCompletionListener(
            final ListenableFuture<T> future,
            final FutureCompletionListener<? super T> listener) {
        if (notificationExecutorService == null) {
            listener.completed(future);
        } else {
            // For future completion listeners, the order does not matter, so we
            // can simply schedule a task for notifying the listener.
            notificationExecutorService.execute(new Runnable() {
                @Override
                public void run() {
                    // If the listener throws, we want to handle this right
                    // here. We do not know which executor service
                    // implementation is used and thus cannot know how it would
                    // handle an exception.
                    try {
                        listener.completed(future);
                    } catch (Throwable t) {
                        errorHandler.handleError(this.getClass(), t,
                                "The completion listener for a future threw an exception.");
                    }
                }
            });
        }
    }

    @Override
    public void setBeaconDetector(CommunicationProcessor communicationProcessor) {
        if (destroyed) {
            throw new IllegalStateException(
                    "Cannot set the beacon detector after the threading strategy has been destroyed.");
        }
        communicationProcessor
                .setCommunicationController(beaconAndResolverThread
                        .getCommunicationController());
    }

    @Override
    public void setChannelNameResolver(
            CommunicationProcessor communicationProcessor) {
        if (destroyed) {
            throw new IllegalStateException(
                    "Cannot set the channel name resolver after the threading strategy has been destroyed.");
        }
        communicationProcessor
                .setCommunicationController(beaconAndResolverThread
                        .getCommunicationController());
    }

    @Override
    public void destroy() {
        destroyed = true;
        beaconAndResolverThread.stopGracefully();
        // We stop all server threads but we do not remove them from the map so
        // that subsequent calls to removeServer do not throw an exception.
        for (CommunicationThread thread : serverThreads.values()) {
            thread.stopGracefully();
        }
        // We know that we have processed all threads that have been started
        // because we set the destroyed flag before iterating over the map. We
        // might have missed threads that have been added and not started, but
        // this is okay. Therefore, we can simply clear the whole map.
        serverThreads.clear();
        if (shutdownNotificationExecutorServiceOnDestruction
                && notificationExecutorService != null) {
            // We use shutdown() instead of shutdownNow(). This avoids problems
            // that might arise when a future completion listener is not
            // notified and thus another thread blocks indefinitely. Usually, a
            // thread waiting on a future to complete should use the future's
            // get() method, but their might be some cases where a more complex
            // (and indirect) setup is used.
            notificationExecutorService.shutdown();
        }
    }

    private void processNotificationQueue(Object synchronizationKey,
            NotificationQueue notificationQueue) {
        Runnable nextTask;
        synchronized (notificationQueue) {
            // If another notification is still running, we cannot run a
            // notification now because this could result in an out-of-order
            // notification. Therefore, we remember that one more task should be
            // executed and return.
            if (notificationQueue.taskIsRunning) {
                ++notificationQueue.executionAttemptsWhileRunning;
                return;
            }
            notificationQueue.taskIsRunning = true;
            nextTask = notificationQueue.tasks.poll();
        }
        // nextTask should never be null because we queue as many tasks in the
        // queue as we schedule with the executor.
        try {
            nextTask.run();
        } finally {
            boolean tasksSkipped;
            synchronized (notificationQueue) {
                notificationQueue.taskIsRunning = false;
                // If tasks where not executed because the task ran was still
                // running, we have to process the corresponding number of tasks
                // now. Otherwise, they would never be executed because the
                // corresponding tasks that have been submitted to the executor
                // service have already finished.
                if (notificationQueue.executionAttemptsWhileRunning > 0) {
                    --notificationQueue.executionAttemptsWhileRunning;
                    tasksSkipped = true;
                } else {
                    tasksSkipped = false;
                }
                // If the queue is empty, we want to remove the entry from the
                // map. We can safely do this because it must still be in the
                // map. No other thread can possibly have removed it as it was
                // not empty before and we are holding the lock.
                if (notificationQueue.tasks.isEmpty()) {
                    notificationQueues.remove(synchronizationKey);
                }
            }
            if (tasksSkipped) {
                processNotificationQueue(synchronizationKey, notificationQueue);
            }
        }
    }

    private void submitNotificationTask(final Object synchronizationKey,
            Runnable notificationTask) {
        NotificationQueue notificationQueue = null;
        while (notificationQueue == null) {
            notificationQueue = notificationQueues.get(synchronizationKey);
            if (notificationQueue == null) {
                notificationQueues.putIfAbsent(synchronizationKey,
                        new NotificationQueue());
            } else {
                synchronized (notificationQueue) {
                    // We have to hold the lock while adding the task. This
                    // ensures that another thread will not remove the queue
                    // from the map at the same time (and also protects the
                    // queue itself from concurrent modifications). If the queue
                    // is not registered in the map any longer, we cannot use it
                    // and have to try again. We compare by identity here: The
                    // queue does not override equals() anyway.
                    if (notificationQueue != notificationQueues
                            .get(synchronizationKey)) {
                        notificationQueue = null;
                    } else {
                        notificationQueue.tasks.add(notificationTask);
                    }
                }
            }
        }
        // As long as the queue is not empty, it will not be removed by
        // another thread, so we can simply use it when the task is finally
        // executed.
        final NotificationQueue finalNotificationQueue = notificationQueue;
        notificationExecutorService.execute(new Runnable() {
            @Override
            public void run() {
                processNotificationQueue(synchronizationKey,
                        finalNotificationQueue);
            }
        });
    }

}
