/*
 * Copyright 2015-2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.resolver;

import java.net.Inet4Address;
import java.net.InetSocketAddress;
import java.nio.charset.Charset;
import java.nio.charset.UnsupportedCharsetException;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import org.apache.commons.lang3.tuple.Pair;

import com.aquenos.epics.jackie.common.exception.ErrorHandler;
import com.aquenos.epics.jackie.common.exception.ExceptionSafeErrorHandler;
import com.aquenos.epics.jackie.common.exception.JavaUtilLoggingErrorHandler;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessConstants;
import com.aquenos.epics.jackie.common.util.EnvironmentUtil;
import com.aquenos.epics.jackie.common.util.HostNameUtil;
import com.aquenos.epics.jackie.common.util.UserNameUtil;

/**
 * Configuration for a {@link ChannelNameResolver}.
 * 
 * @author Sebastian Marsching
 */
public final class ChannelNameResolverConfiguration {

    private Charset charset;
    private double echoInterval;
    private ErrorHandler errorHandler;
    private int defaultServerPort;
    private HashSet<InetSocketAddress> tcpAddresses = new HashSet<InetSocketAddress>();
    private HashSet<InetSocketAddress> udpAddresses = new HashSet<InetSocketAddress>();
    private boolean autoDetermineAddresses;
    private double maximumSearchPeriod;
    private int multicastTimeToLive;
    private String hostName;
    private String userName;

    /**
     * Creates a configuration for a channel-name resolver using environment
     * variables. The error handler is initialized with a
     * {@link JavaUtilLoggingErrorHandler}. The <code>EPICS_CA_ADDR_LIST</code>,
     * <code>EPICS_CA_AUTO_ADDR_LIST</code>, <code>EPICS_CA_NAME_SERVERS</code>,
     * <code>EPICS_CA_SERVER_PORT</code>,
     * <code>EPICS_CA_MAX_SEARCH_PERIOD</code>, <code>EPICS_CA_CONN_TMO</code>
     * and <code>EPICS_CA_MCAST_TTL</code> environment variables are used for
     * initializing the other configuration options. They have exactly the same
     * meanings and default values as specified by the Channel Access Reference
     * Manual.
     */
    public ChannelNameResolverConfiguration() {
        this(null, null, null, null, null, null, null, null, null, null, null);
    }

    /**
     * Creates a channel-name resolver configuration using the specified
     * options. Each option, that is <code>null</code> is initialized from its
     * corresponding environment variable, as the constructor
     * {@link #ChannelNameResolverConfiguration()} does.
     * 
     * @param charset
     *            character set used to encode and decode channel names. If this
     *            parameter is <code>null</code> the UTF-8 character set is
     *            used. If the UTF-8 character set is not available on this
     *            platform, the platform's default character set is used.
     * @param errorHandler
     *            error handler which is used to log errors that are encountered
     *            by the channel-name resolver or <code>null</code> if a
     *            {@link JavaUtilLoggingErrorHandler} shall be used.
     * @param defaultServerPort
     *            default port to assume for servers if no port is specified
     *            explicitly or <code>null</code> to use
     *            <code>EPICS_CA_SERVER_PORT</code>.
     * @param tcpAddresses
     *            set of IP-address port pairs of channel name servers that
     *            should be queried via TCP or <code>null</code> to use
     *            <code>EPICS_CA_NAME_SERVERS</code>.
     * @param udpAddresses
     *            set of IP-address port pairs of channel name servers or
     *            broadcast addresses that should be queried via UDP or
     *            <code>null</code> to use <code>EPICS_CA_ADDR_LIST</code>.
     * @param autoDetermineAddresses
     *            <code>true</code> if the broadcast addresses of all local
     *            interfaces should be added to the set of addresses that are
     *            queried via UDP, <code>false</code> if those addresses should
     *            not be added, and <code>null</code> if
     *            <code>EPICS_CA_AUTO_ADDR_LIST</code> should be used.
     * @param maximumSearchPeriod
     *            interval of the longest search period (in seconds) or
     *            <code>null</code> if <code>EPICS_CA_MAX_SEARCH_PERIOD</code>
     *            should be used. The minimum value for this setting is 60. If
     *            it is below 60, it is silently set to 60.
     * @throws IllegalArgumentException
     *             if <code>defaultServerPort</code> is specified and zero or an
     *             invalid port number and if <code>tcpAddresses</code> or
     *             <code>udpAddresses</code> contains an entry with a zero or
     *             invalid port number.
     * @deprecated Since 2.0.0. Use
     *             {@link #ChannelNameResolverConfiguration(Charset, ErrorHandler, String, String, Integer, Set, Set, Boolean, Double, Double, Integer)}
     *             instead.
     */
    public ChannelNameResolverConfiguration(Charset charset,
            ErrorHandler errorHandler, Integer defaultServerPort,
            Set<Pair<Inet4Address, Integer>> tcpAddresses,
            Set<Pair<Inet4Address, Integer>> udpAddresses,
            Boolean autoDetermineAddresses, Double maximumSearchPeriod) {
        this(charset, errorHandler, null, null, defaultServerPort, tcpAddresses,
                udpAddresses, autoDetermineAddresses, maximumSearchPeriod, null,
                null);
    }

    /**
     * Creates a channel-name resolver configuration using the specified
     * options. Each option, that is <code>null</code> is initialized from its
     * corresponding environment variable, as the constructor
     * {@link #ChannelNameResolverConfiguration()} does.
     * 
     * @param charset
     *            character set used to encode and decode channel names. If this
     *            parameter is <code>null</code> the UTF-8 character set is
     *            used. If the UTF-8 character set is not available on this
     *            platform, the platform's default character set is used.
     * @param errorHandler
     *            error handler which is used to log errors that are encountered
     *            by the channel-name resolver or <code>null</code> if a
     *            {@link JavaUtilLoggingErrorHandler} shall be used.
     * @param hostName
     *            host name identifying the client computer when connecting to a
     *            server. If <code>null</code> the host name is determined
     *            automatically using the {@link HostNameUtil#getHostName()}
     *            method. If the host name cannot be determined automatically,
     *            the special string "&lt;unknown host&gt;" is used.
     * @param userName
     *            user name identifying the user when connecting to a server. If
     *            <code>null</code> the user name is determined automatically
     *            using the {@link UserNameUtil#getUserName()} method. If the
     *            user name cannot be determined automatically, the empty string
     *            is used.
     * @param defaultServerPort
     *            default port to assume for servers if no port is specified
     *            explicitly or <code>null</code> to use
     *            <code>EPICS_CA_SERVER_PORT</code>.
     * @param tcpAddresses
     *            set of IP-address port pairs of channel name servers that
     *            should be queried via TCP or <code>null</code> to use
     *            <code>EPICS_CA_NAME_SERVERS</code>.
     * @param udpAddresses
     *            set of IP-address port pairs of channel name servers or
     *            broadcast addresses that should be queried via UDP or
     *            <code>null</code> to use <code>EPICS_CA_ADDR_LIST</code>.
     * @param autoDetermineAddresses
     *            <code>true</code> if the broadcast addresses of all local
     *            interfaces should be added to the set of addresses that are
     *            queried via UDP, <code>false</code> if those addresses should
     *            not be added, and <code>null</code> if
     *            <code>EPICS_CA_AUTO_ADDR_LIST</code> should be used.
     * @param maximumSearchPeriod
     *            interval of the longest search period (in seconds) or
     *            <code>null</code> if <code>EPICS_CA_MAX_SEARCH_PERIOD</code>
     *            should be used. The minimum value for this setting is 60. If
     *            it is below 60, it is silently set to 60.
     * @param echoInterval
     *            interval between sending echo requests to a TCP-based
     *            name-server in seconds. If <code>null</code>, the
     *            <code>EPICS_CA_CONN_TMO</code> environment variable is used.
     *            If the environment variable is not set or its value is
     *            invalid, a default value of 30 seconds is used.
     * @param multicastTimeToLive
     *            TTL to be used for UDP search packets that are sent to
     *            multicast addresses. If <code>null</code> the
     *            <code>EPICS_CA_MCAST_TTL</code> environment variable is used.
     *            If the environment variable is not set or its value is
     *            invalid, a default value of 1 is used. The minimum value for
     *            this setting is 1 and the maximum value is 255.
     * @throws IllegalArgumentException
     *             if <code>defaultServerPort</code> is specified and zero or an
     *             invalid port number, <code>tcpAddresses</code> or
     *             <code>udpAddresses</code> contains an entry with a zero or
     *             invalid port number, or <code>multicastTimeToLive</code> is
     *             out of range.
     * @since EPICS Jackie 2.0.0
     */
    public ChannelNameResolverConfiguration(Charset charset,
            ErrorHandler errorHandler, String hostName, String userName,
            Integer defaultServerPort,
            Set<Pair<Inet4Address, Integer>> tcpAddresses,
            Set<Pair<Inet4Address, Integer>> udpAddresses,
            Boolean autoDetermineAddresses, Double maximumSearchPeriod,
            Double echoInterval, Integer multicastTimeToLive) {
        if (charset != null) {
            this.charset = charset;
        } else {
            try {
                this.charset = Charset.forName("UTF-8");
            } catch (UnsupportedCharsetException e) {
                this.charset = Charset.defaultCharset();
            }
        }
        if (errorHandler != null) {
            this.errorHandler = new ExceptionSafeErrorHandler(errorHandler);
        } else {
            this.errorHandler = new ExceptionSafeErrorHandler(
                    new JavaUtilLoggingErrorHandler());
        }
        if (defaultServerPort != null) {
            this.defaultServerPort = defaultServerPort;
            if (defaultServerPort <= 0 || defaultServerPort > 65535) {
                throw new IllegalArgumentException(
                        "Invalid default server port: "
                                + this.defaultServerPort);
            }
        } else {
            initializeDefaultServerPortFromEnvironment();
        }
        if (tcpAddresses != null) {
            for (Pair<Inet4Address, Integer> addressPair : tcpAddresses) {
                if (addressPair.getLeft() == null) {
                    throw new NullPointerException();
                }
                int port = addressPair.getRight();
                if (port <= 0 || port > 65535) {
                    throw new IllegalArgumentException(
                            "Invalid entry in TCP addresses: "
                                    + addressPair.getLeft().getHostAddress()
                                    + ":" + port);
                }
                this.tcpAddresses.add(new InetSocketAddress(
                        addressPair.getLeft(), addressPair.getRight()));
            }
        } else {
            initializeTcpAddressesFromEnvironment();
        }
        if (udpAddresses != null) {
            for (Pair<Inet4Address, Integer> addressPair : udpAddresses) {
                if (addressPair.getLeft() == null) {
                    throw new NullPointerException();
                }
                int port = addressPair.getRight();
                if (port <= 0 || port > 65535) {
                    throw new IllegalArgumentException(
                            "Invalid entry in UDP addresses: "
                                    + addressPair.getLeft().getHostAddress()
                                    + ":" + port);
                }
                this.udpAddresses.add(new InetSocketAddress(
                        addressPair.getLeft(), addressPair.getRight()));
            }
        } else {
            initializeUdpAddressesFromEnvironment();
        }
        if (autoDetermineAddresses != null) {
            this.autoDetermineAddresses = autoDetermineAddresses;
        } else {
            initializeAutoDetermineAddressesFromEnvironment();
        }
        if (maximumSearchPeriod != null) {
            this.maximumSearchPeriod = maximumSearchPeriod;
        } else {
            initializeMaximumSearchPeriodFromEnvironment();
        }
        if (this.maximumSearchPeriod < 60.0
                || Double.isNaN(this.maximumSearchPeriod)) {
            this.maximumSearchPeriod = 60.0;
        }
        if (echoInterval != null) {
            if (echoInterval < 0.1 || Double.isInfinite(echoInterval)
                    || Double.isNaN(echoInterval)) {
                throw new IllegalArgumentException(
                        "The echo interval must be at least 0.1 s.");
            }
            this.echoInterval = echoInterval;
        } else {
            initializeEchoIntervalFromEnvironment();
        }
        if (multicastTimeToLive != null) {
            if (multicastTimeToLive < 1 || multicastTimeToLive > 255) {
                throw new IllegalArgumentException(
                        "The multicast TTL must be in the range 1 <= TTL <= 255.");
            }
            this.multicastTimeToLive = multicastTimeToLive;
        } else {
            initializeMulticastTimeToLiveFromEnvironment();
        }
        if (hostName != null) {
            this.hostName = hostName;
        } else {
            initializeHostNameFromEnvironment();
        }
        if (userName != null) {
            this.userName = userName;
        } else {
            initializeUserNameFromEnvironment();
        }
    }

    private void initializeEchoIntervalFromEnvironment() {
        this.echoInterval = EnvironmentUtil
                .doubleFromEnvironment("EPICS_CA_CONN_TMO", 30.0);
        if (this.echoInterval < 0.1) {
            this.echoInterval = 0.1;
        }
    }

    private void initializeDefaultServerPortFromEnvironment() {
        defaultServerPort = EnvironmentUtil.portNumberFromEnvironment(
                "EPICS_CA_SERVER_PORT",
                ChannelAccessConstants.DEFAULT_SERVER_PORT, false);
    }

    private void initializeTcpAddressesFromEnvironment() {
        for (Pair<Inet4Address, Integer> address : EnvironmentUtil
                .addressAndPortListFromEnvironment("EPICS_CA_NAME_SERVERS",
                        defaultServerPort)) {
            // We cannot use a port number of zero as a destination port.
            if (address.getRight() != 0) {
                tcpAddresses.add(new InetSocketAddress(address.getLeft(),
                        address.getRight()));
            }
        }
    }

    private void initializeUdpAddressesFromEnvironment() {
        for (Pair<Inet4Address, Integer> address : EnvironmentUtil
                .addressAndPortListFromEnvironment("EPICS_CA_ADDR_LIST",
                        defaultServerPort)) {
            // We cannot use a port number of zero as a destination port.
            if (address.getRight() != 0) {
                udpAddresses.add(new InetSocketAddress(address.getLeft(),
                        address.getRight()));
            }
        }
    }

    private void initializeAutoDetermineAddressesFromEnvironment() {
        autoDetermineAddresses = EnvironmentUtil
                .booleanFromEnvironment("EPICS_CA_AUTO_ADDR_LIST", true);
    }

    private void initializeMaximumSearchPeriodFromEnvironment() {
        maximumSearchPeriod = EnvironmentUtil
                .doubleFromEnvironment("EPICS_CA_MAX_SEARCH_PERIOD", 300.0);
    }

    private void initializeMulticastTimeToLiveFromEnvironment() {
        multicastTimeToLive = EnvironmentUtil
                .integerFromEnvironment("EPICS_CA_MCAST_TTL", 1, 1, 255);
    }

    private void initializeHostNameFromEnvironment() {
        this.hostName = HostNameUtil.getHostName();
        if (this.hostName == null) {
            this.hostName = "<unknown host>";
        }
    }

    private void initializeUserNameFromEnvironment() {
        this.userName = UserNameUtil.getUserName();
        if (this.userName == null) {
            this.userName = "";
        }
    }

    /**
     * Returns the character set that should be used for encoding and decoding
     * channel names.
     * 
     * @return character set to be used for channel names.
     */
    public Charset getCharset() {
        return charset;
    }

    /**
     * <p>
     * Returns the interval between two echo requests.
     * </p>
     * 
     * <p>
     * Echo requests are sent to a TCP-based name-server in order to ensure that
     * the connection is still working. If a beacon packet from the same server
     * is received, this is considered proof that the network connection is okay
     * and thus no echo request is sent. For this reason, this setting should be
     * chosen to be at least two times as large as the max. interval between
     * sending two beacons.
     * </p>
     * 
     * <p>
     * Unlike for channel-servers, the echo mechanism does not affect the
     * connected state of any name-server. However, sending echo requests when
     * no beacons have been received still makes sense because actually sending
     * some data might help the operating system in detecting that the TCP
     * connection has been interrupted.
     * </p>
     * 
     * <p>
     * The minimum for this setting is 0.1 seconds.
     * </p>
     * 
     * @return echo interval (in seconds).
     */
    public double getEchoInterval() {
        return echoInterval;
    }

    /**
     * Returns the error handler that should be used by the channel-name
     * resolver. This configuration object ensures that the error handler
     * returned will never throw an exception by wrapping it in a
     * {@link ExceptionSafeErrorHandler}.
     * 
     * @return error handler that is exception safe.
     */
    public ErrorHandler getErrorHandler() {
        return errorHandler;
    }

    /**
     * Returns the TCP and UDP port that should be used when connecting to a
     * server for name resolution and no port has been specified for this server
     * explicitly.
     * 
     * @return default server port for name resolution.
     */
    public int getDefaultServerPort() {
        return defaultServerPort;
    }

    /**
     * Returns socket addresses of server addresses that should be contacted via
     * TCP for name resolution.
     * 
     * @return addresses of TCP name servers.
     */
    public Set<InetSocketAddress> getTcpAddresses() {
        return Collections.unmodifiableSet(tcpAddresses);
    }

    /**
     * Returns socket addresses of server addresses and broadcast addresses that
     * should be contacted via UDP for name resolution.
     * 
     * @return addresses of UDP name servers and broadcast addresses.
     */
    public Set<InetSocketAddress> getUdpAddresses() {
        return Collections.unmodifiableSet(udpAddresses);
    }

    /**
     * Returns whether the broadcast addresses for the local interfaces should
     * also be contacted via UDP, even if they are not explicitly listed in the
     * set returned by {@link #getUdpAddresses()}.
     * 
     * @return <code>true</code> if the broadcast addresses of all local
     *         interfaces should be used for name resolution, <code>false</code>
     *         if only the explicitly configured address should be used.
     */
    public boolean isAutoDetermineAddresses() {
        return autoDetermineAddresses;
    }

    /**
     * Returns the maximum search period (in seconds). The maximum search period
     * defines the longest time interval between two attempts to resolve a
     * channel name. However, internally a slightly different time might be used
     * so this value rather defines an order of magnitude. The minimum for this
     * setting is 60. The effective maximum is implementation dependent. For a
     * typical implementation, it is approximately 2000.
     * 
     * @return maximum search period (at least 60).
     */
    public double getMaximumSearchPeriod() {
        return maximumSearchPeriod;
    }

    /**
     * Returns the multicast TTL. This TTL is used when sending UDP packets to a
     * multicast address. Due to API restrictions, the TTL cannot be set when
     * running on Java 6. In this case, the multicast TTL is not actually set
     * and some platform-specific default is used instead.
     * 
     * @return multicast TTL for name-search UDP packets.
     */
    public int getMulticastTimeToLive() {
        return multicastTimeToLive;
    }

    /**
     * Returns the host name that shall be used as the identification of the
     * local machine when connecting to a Channel Access server.
     * 
     * @return host name identifying the computer the client is running on.
     */
    public String getHostName() {
        return hostName;
    }

    /**
     * Returns the user name that shall be used as the identification of the
     * user when connecting to a Channel Access server.
     * 
     * @return user name identifying the user that is using the client.
     */
    public String getUserName() {
        return userName;
    }

}
