/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import org.apache.commons.lang3.builder.HashCodeBuilder;

/**
 * <p>
 * Mask used for the subscription to a channel. The mask defines which events
 * are sent to the client. An event is only sent to the client if at least one
 * of the bits in the event's event mask is also set in the event mask the
 * client set when creating the subscription.
 * </p>
 * 
 * <p>
 * Instances of this class are constant. Therefore, methods that modify the mask
 * do not modify the instance they are called one. Instead, they return a new
 * instance with the modification (like {@link String}s do).
 * </p>
 * 
 * @author Sebastian Marsching
 */
public final class ChannelAccessEventMask {

    private static final ChannelAccessEventMask[] ALL_MASKS;
    static {
        ALL_MASKS = new ChannelAccessEventMask[16];
        for (int i = 0; i < ALL_MASKS.length; i++) {
            ALL_MASKS[i] = new ChannelAccessEventMask(i);
        }
    }

    /**
     * Mask that has no flag set. A subscription with such a mask will not
     * receive any events and an event with this mask is not sent to any client.
     * Therefore, this mask is only useful as an empty start for creating a
     * custom mask.
     */
    public static final ChannelAccessEventMask DBE_NONE = ALL_MASKS[0];

    /**
     * Mask that has only the <code>DBE_VALUE</code> flag set. An event with
     * this mask is triggered for every significant value change. The dead-band
     * for significant value changes is configured on the server side.
     */
    public static final ChannelAccessEventMask DBE_VALUE = ALL_MASKS[1];

    /**
     * Mask that has only the <code>DBE_ARCHIVE</code> flag set. An event with
     * this mask is triggered for every archive-significant value change. The
     * dead-band for archive-significant value changes is configured on the
     * server side. Typically, the dead-band for such events is at least as
     * large as the one for <code>DBE_VALUE</code> events.
     */
    public static final ChannelAccessEventMask DBE_ARCHIVE = ALL_MASKS[2];

    /**
     * Mask that has only the <code>DBE_ALARM</code> flag set. An event with
     * this mask is triggered when a channel's alarm state changes.
     */
    public static final ChannelAccessEventMask DBE_ALARM = ALL_MASKS[4];

    /**
     * Mask that has only the <code>DBE_PROPERTY</code> flag set. An event with
     * this mask is triggered when a property of a channel (e.g. display limits,
     * engineering units, enum labels, etc.) changes.
     */
    public static final ChannelAccessEventMask DBE_PROPERTY = ALL_MASKS[8];

    /**
     * Mask that has all flags (<code>DBE_VALUE</code>, <code>DBE_ARCHIVE</code>
     * , <code>DBE_ALARM</code>, and <code>DBE_PROPERTY</code>) set.
     */
    public static final ChannelAccessEventMask DBE_ALL = ALL_MASKS[15];

    private final int eventMask;
    private final int hashCode;

    private ChannelAccessEventMask(int eventMask) {
        this.eventMask = eventMask;
        this.hashCode = new HashCodeBuilder(89, 23).append(eventMask)
                .toHashCode();
    }

    /**
     * Returns the mask that is the result of combining this mask and the
     * <code>other</code> mask using a binary <em>and</em> operation. This means
     * that the resulting mask has all flags set that are set in both this
     * instance and the other instance.
     * 
     * @param other
     *            mask this mask should be combined with.
     * @return mask that is the result of a binary and of this mask and the
     *         specified mask.
     */
    public ChannelAccessEventMask and(ChannelAccessEventMask other) {
        return ALL_MASKS[this.eventMask & other.eventMask];
    }

    /**
     * Returns the mask that is the result of combining this mask and the
     * <code>other</code> mask using a binary <em>or</em> operation. This means
     * that the resulting mask has all flags set that are set in this instance
     * or the other instance.
     * 
     * @param other
     *            mask this mask should be combined with.
     * @return mask that is the result of a binary or of this mask and the
     *         specified mask.
     */
    public ChannelAccessEventMask or(ChannelAccessEventMask other) {
        return ALL_MASKS[this.eventMask | other.eventMask];
    }

    /**
     * Returns the mask that is the inverse of this mask. The inverse has all
     * flags set that this mask does not have set and has all the flags not set
     * that this mask has set. This is equivalent to applying a binary
     * <em>not</em> operation.
     * 
     * @return inverse of this mask.
     */
    public ChannelAccessEventMask not() {
        return ALL_MASKS[(~eventMask) & 0x0f];
    }

    /**
     * Returns whether any of the flags in this mask are set.
     * 
     * @return <code>true</code> if at least one flag in this mask is set,
     *         <code>false</code> if none of the flags is set (this mask is
     *         equal to {@link #DBE_NONE}).
     */
    public boolean isAny() {
        return eventMask != 0;
    }

    /**
     * Returns whether this mask has the the {@link #DBE_VALUE} flag set. This
     * is the same as using <code>this.and(DBE_VALUE).isAny()</code>.
     * 
     * @return <code>true</code> if this mask has the <code>DBE_VALUE</code>
     *         flag set.
     */
    public boolean isValue() {
        return this.and(DBE_VALUE).isAny();
    }

    /**
     * Sets or clears the {@link #DBE_VALUE} flag. This does not modify this
     * mask instance. The modified mask is returned instead.
     * 
     * @param flag
     *            <code>true</code> if the <code>DBE_VALUE</code> flag should be
     *            set, <code>false</code> if it should be cleared.
     * @return the mask that is the same as this mask but has the
     *         <code>DBE_VALUE</code> flag set or cleared.
     */
    public ChannelAccessEventMask setValue(boolean flag) {
        if (flag) {
            return this.or(DBE_VALUE);
        } else {
            return this.and(DBE_VALUE.not());
        }
    }

    /**
     * Returns whether this mask has the the {@link #DBE_ARCHIVE} flag set. This
     * is the same as using <code>this.and(DBE_ARCHIVE).isAny()</code>.
     * 
     * @return <code>true</code> if this mask has the <code>DBE_ARCHIVE</code>
     *         flag set.
     */
    public boolean isArchive() {
        return this.and(DBE_ARCHIVE).isAny();
    }

    /**
     * Sets or clears the {@link #DBE_ARCHIVE} flag. This does not modify this
     * mask instance. The modified mask is returned instead.
     * 
     * @param flag
     *            <code>true</code> if the <code>DBE_ARCHIVE</code> flag should
     *            be set, <code>false</code> if it should be cleared.
     * @return the mask that is the same as this mask but has the
     *         <code>DBE_ARCHIVE</code> flag set or cleared.
     */
    public ChannelAccessEventMask setArchive(boolean flag) {
        if (flag) {
            return this.or(DBE_ARCHIVE);
        } else {
            return this.and(DBE_ARCHIVE.not());
        }
    }

    /**
     * Returns whether this mask has the the {@link #DBE_ALARM} flag set. This
     * is the same as using <code>this.and(DBE_ALARM).isAny()</code>.
     * 
     * @return <code>true</code> if this mask has the <code>DBE_ALARM</code>
     *         flag set.
     */
    public boolean isAlarm() {
        return this.and(DBE_ALARM).isAny();
    }

    /**
     * Sets or clears the {@link #DBE_ALARM} flag. This does not modify this
     * mask instance. The modified mask is returned instead.
     * 
     * @param flag
     *            <code>true</code> if the <code>DBE_ALARM</code> flag should be
     *            set, <code>false</code> if it should be cleared.
     * @return the mask that is the same as this mask but has the
     *         <code>DBE_ALARM</code> flag set or cleared.
     */
    public ChannelAccessEventMask setAlarm(boolean flag) {
        if (flag) {
            return this.or(DBE_ALARM);
        } else {
            return this.and(DBE_ALARM.not());
        }
    }

    /**
     * Returns whether this mask has the the {@link #DBE_PROPERTY} flag set.
     * This is the same as using <code>this.and(DBE_PROPERTY).isAny()</code>.
     * 
     * @return <code>true</code> if this mask has the <code>DBE_PROPERTY</code>
     *         flag set.
     */
    public boolean isProperty() {
        return this.and(DBE_PROPERTY).isAny();
    }

    /**
     * Sets or clears the {@link #DBE_PROPERTY} flag. This does not modify this
     * mask instance. The modified mask is returned instead.
     * 
     * @param flag
     *            <code>true</code> if the <code>DBE_PROPERTY</code> flag should
     *            be set, <code>false</code> if it should be cleared.
     * @return the mask that is the same as this mask but has the
     *         <code>DBE_PROPERTY</code> flag set or cleared.
     */
    public ChannelAccessEventMask setProperty(boolean flag) {
        if (flag) {
            return this.or(DBE_PROPERTY);
        } else {
            return this.and(DBE_PROPERTY.not());
        }
    }

    /**
     * Returns the numeric representation for this mask. This numeric
     * representation is suitable for transferring the mask over a network.
     * 
     * @return numeric representation of this mask.
     */
    public short toNumericMask() {
        return (short) eventMask;
    }

    /**
     * Tells whether the specified object represents the same event mask. Event
     * masks are considered to be equal if and only if all of their mask bits
     * are identical.
     * 
     * @param obj
     *            reference object with which this object is compared.
     * @return <code>true</code> if <code>obj</code> is an event mask that has
     *         exactly the same bits set as this mask, <code>false</code>
     *         otherwise.
     */
    @Override
    public boolean equals(Object obj) {
        // We know that each possible event mask object exists exactly once.
        // Therefore, we can use object identity as equality.
        return this == obj;
    }

    /**
     * <p>
     * Returns a hash code value for this message. This method is supported for
     * the benefit of hash tables. The hash code is calculated based on the
     * combination of all mask bits.
     * </p>
     * 
     * <p>
     * This means that two identical masks always have the same hash code. The
     * calculation of the hash code is designed in a way so that two different
     * masks are very likely to have different hash codes, but this is not
     * strictly guaranteed.
     * </p>
     * 
     * @return hash code suitable for use by a hash table.
     */
    @Override
    public int hashCode() {
        return hashCode;
    }

    /**
     * Returns a string representation of this event mask. The string
     * representation is generated by adding a text with the name (e.g.
     * <code>DBE_VALUE</code>) of each bit that is set in this mask, separating
     * the names by the pipe character("|") if more than one bit is set.
     * 
     * @return string representation of this event mask.
     */
    @Override
    public String toString() {
        if (eventMask == 0) {
            return "DBE_NONE";
        }
        StringBuilder sb = new StringBuilder();
        boolean first = true;
        if (isValue()) {
            sb.append("DBE_VALUE");
            first = false;
        }
        if (isArchive()) {
            if (!first) {
                sb.append('|');
            }
            sb.append("DBE_ARCHIVE");
            first = false;
        }
        if (isAlarm()) {
            if (!first) {
                sb.append('|');
            }
            sb.append("DBE_ALARM");
            first = false;
        }
        if (isProperty()) {
            if (!first) {
                sb.append('|');
            }
            sb.append("DBE_PROPERTY");
            first = false;
        }
        return sb.toString();
    }

    /**
     * Returns the mask that is represented by the specified numeric mask. Only
     * the least significant four bits of the numeric mask are considered. This
     * method is suitable for getting the mask instance for a mask that has been
     * transferred over the network.
     * 
     * @param mask
     *            numeric mask (e.g. received over the network). Only the least
     *            significant four bits are used.
     * @return mask instance that represents the specified mask.
     * @see #toNumericMask()
     */
    public static ChannelAccessEventMask forNumericMask(short mask) {
        return ALL_MASKS[mask & 0x0f];
    }

}
