/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;

/**
 * Representation of a <code>CA_PROTO_READ_BUILD</code> message. This message
 * can only be received, not sent. It is only supported for compatibility with
 * legacy clients. A server that receives this message should ignore it.
 * 
 * @author Sebastian Marsching
 * @see ChannelAccessCommand#CA_PROTO_READ_BUILD
 */
public final class ChannelAccessReadBuildMessage extends ChannelAccessMessage {

    private ChannelAccessReadBuildMessage(
            ChannelAccessMessageHeader messageHeader) {
        super(ChannelAccessCommand.CA_PROTO_READ_BUILD, messageHeader);
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        // This method should never be called because we already filter this
        // type of message in the ChannelAccessMessageCodec. If this method gets
        // called, something must be wrong.
        throw new UnsupportedOperationException(
                "The CA_PROTO_READ_BUILD message cannot be sent by this implementation.");
    }

    // This message has no state, so there is no need to override equals and
    // hashCode from the super class.

    @Override
    protected void verify(ChannelAccessVersion version, int maxPayloadSize,
            Charset charset) {
        // This method should never be called because we already filter this
        // type of message in the ChannelAccessMessageCodec. If this method gets
        // called, something must be wrong.
        throw new UnsupportedOperationException(
                "The CA_PROTO_READ_BUILD message cannot be sent by this implementation.");
    }

    @Override
    public String toString() {
        return command.toString();
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @param headerOnly
     *            if <code>true</code> this method will not try to read the
     *            payload but just read the header. This is useful for reading a
     *            message that is sent in the payload of a
     *            {@link ChannelAccessErrorMessage}. Such a message will have
     *            all headers but no payload (even so the payload size might be
     *            non-zero).
     * @return message constructed from the data in the header and the payload
     *         read from the byte source (if <code>headerOnly</code> is
     *         <code>false</code>).
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete message (including he payload if
     *             <code>headerOnly</code> is <code>false</code>).
     */
    protected static ChannelAccessReadBuildMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource,
            boolean headerOnly) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_READ_BUILD
                .getCommandNumber());
        if (!headerOnly) {
            byteSource.skip(messageHeader.getPayloadSize());
        }
        return new ChannelAccessReadBuildMessage(messageHeader);
    }

}
