/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.StringEscapeUtils;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.apache.commons.lang3.tuple.Pair;

import com.aquenos.epics.jackie.common.exception.SerializedChannelNameTooLargeException;
import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.util.NullTerminatedStringUtil;

/**
 * Representation of a <code>CA_PROTO_SEARCH</code> message that is sent from a
 * client to a UDP or TCP server in order to find the location of a channel.
 * 
 * @author Sebastian Marsching
 * @see ChannelAccessCommand#CA_PROTO_SEARCH
 */
public final class ChannelAccessSearchClientMessage extends
        ChannelAccessSearchMessage {

    private final static short DO_REPLY = 10;
    private final static short DO_NOT_REPLY = 5;

    private int cid;
    private String channelName;
    private boolean doReply;
    private ChannelAccessVersion version;

    /**
     * Creates a new <code>CA_PROTO_SEARCH</code> message that can be sent from
     * a client to a UDP or TCP server.
     * 
     * @param cid
     *            client-side identifier for this channel.
     * @param version
     *            Channel Access protocol version supported by the client.
     * @param doReply
     *            <code>true</code> if the client also wants to be informed if
     *            the requested channel could not be resolved,
     *            <code>false</code> if the client only wants to receive
     *            messages with the channel location. This flag might be ignored
     *            by some servers, which will have the same effect as if it were
     *            set to <code>false</code>.
     * @param channelName
     *            name of the channel the client is looking for.
     */
    public ChannelAccessSearchClientMessage(int cid,
            ChannelAccessVersion version, boolean doReply, String channelName) {
        super();

        if (version == null) {
            throw new NullPointerException(
                    "Channel Access version must not be null.");
        }
        if (channelName == null) {
            throw new NullPointerException("Channel name must not be null.");
        }

        this.cid = cid;
        this.version = version;
        this.doReply = doReply;
        this.channelName = channelName;
    }

    private ChannelAccessSearchClientMessage(
            ChannelAccessMessageHeader messageHeader, int cid,
            ChannelAccessVersion version, boolean doReply, String channelName) {
        super(messageHeader);

        this.cid = cid;
        this.version = version;
        this.doReply = doReply;
        this.channelName = channelName;
    }

    /**
     * Returns the channel ID that is used by the client to identify the
     * channel. This integer may be zero or negative.
     * 
     * @return client-side ID for the channel.
     */
    public int getChannelCID() {
        return cid;
    }

    /**
     * Returns the channel name that the client is looking for.
     * 
     * @return channel name. Might be <code>null</code> if this message instance
     *         has been reconstructed without its payload (as part of the
     *         payload of an error message).
     */
    public String getChannelName() {
        return channelName;
    }

    /**
     * Returns the "do reply" flag. This flag indicates that the client wants to
     * be informed by the server if the server does not know the channel the
     * client is looking for. By default, a server will only reply to a search
     * request if it knows the channel the client is looking for. Even if this
     * flag is set, some servers might ignore it and not reply to search
     * requests that they cannot provide an answer for.
     * 
     * @return <code>true</code> if the client wants to be informed by a server
     *         if the channel is not known, <code>false</code> if it only wants
     *         to be informed if the channel was found.
     */
    public boolean isDoReply() {
        return doReply;
    }

    /**
     * Returns the Channel Access version supported by this client. A server can
     * use this information in order to decide whether to reply to a search
     * request. For example, a server might not reply to a search request if the
     * protocol version supported by the client is too old for the server.
     * 
     * @return Channel Access version supported by the client.
     */
    public ChannelAccessVersion getVersion() {
        return version;
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessSearchClientMessage other = (ChannelAccessSearchClientMessage) obj;
        return new EqualsBuilder().append(this.cid, other.cid)
                .append(this.channelName, other.channelName)
                .append(this.doReply, other.doReply)
                .append(this.version, other.version).isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode()).append(cid)
                .append(channelName).append(doReply).append(version)
                .toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("cid=");
        sb.append(cid);
        sb.append(", channelName=");
        if (channelName != null) {
            sb.append('"');
            sb.append(StringEscapeUtils.escapeJava(channelName));
            sb.append('"');
        } else {
            sb.append("null");
        }
        sb.append(", doReply=");
        sb.append(doReply);
        sb.append(", version=");
        sb.append(version);
        sb.append('}');
        return sb.toString();
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        if (channelName == null) {
            throw new UnsupportedOperationException(
                    "Cannot serialize the message payload for a message that has been constructed without the payload.");
        }
        Pair<byte[], Integer> channelNameBytesWithLength = NullTerminatedStringUtil
                .stringToSizeLimitedNullTerminatedBytesOrNull(channelName,
                        ChannelAccessConstants.MAX_CHANNEL_NAME_BYTES_LENGTH,
                        ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT, charset);
        if (channelNameBytesWithLength == null) {
            throw new SerializedChannelNameTooLargeException(
                    "The serialized form of the channel name must not be longer than "
                            + ChannelAccessConstants.MAX_CHANNEL_NAME_BYTES_LENGTH
                            + " bytes.");
        }
        byte[] channelNameAsBytes = channelNameBytesWithLength.getLeft();
        int channelNameLength = channelNameBytesWithLength.getRight();
        // We have to make sure that the channel name is null-terminated and
        // aligned to 8 bytes.
        serializeHeader(byteSink, channelNameLength, doReply ? DO_REPLY
                : DO_NOT_REPLY, this.version.getMinorVersionAsShort(), cid,
                cid, version, maxPayloadSize);
        byteSink.putByteArray(channelNameAsBytes, 0, channelNameLength);
    }

    @Override
    protected void verify(ChannelAccessVersion version, int maxPayloadSize,
            Charset charset) {
        if (channelName == null) {
            throw new UnsupportedOperationException(
                    "Cannot serialize the message payload for a message that has been constructed without the payload.");
        }
        Pair<byte[], Integer> channelNameBytesWithLength = NullTerminatedStringUtil
                .stringToSizeLimitedNullTerminatedBytesOrNull(channelName,
                        ChannelAccessConstants.MAX_CHANNEL_NAME_BYTES_LENGTH,
                        ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT, charset);
        if (channelNameBytesWithLength == null) {
            throw new SerializedChannelNameTooLargeException(
                    "The serialized form of the channel name must not be longer than "
                            + ChannelAccessConstants.MAX_CHANNEL_NAME_BYTES_LENGTH
                            + " bytes.");
        }
        // We can skip the verification of the header.
        // MAX_CHANNEL_NAME_BYTES_LENGTH is shorter than the smallest maximum
        // payload size.
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @param headerOnly
     *            if <code>true</code> this method will not try to read the
     *            payload but just read the header. This is useful for reading a
     *            message that is sent in the payload of a
     *            {@link ChannelAccessErrorMessage}. Such a message will have
     *            all headers but no payload (even so the payload size might be
     *            non-zero).
     * @param charset
     *            encoding to use for deserializing the channel name. Usually,
     *            channel names should consist of ASCII characters only, however
     *            this is not strictly enforced by the protocol, so the encoding
     *            might matter.
     * @return message constructed from the data in the header and the payload
     *         read from the byte source (if <code>headerOnly</code> is
     *         <code>false</code>).
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete payload.
     */
    protected static ChannelAccessSearchClientMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource,
            boolean headerOnly, Charset charset) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_SEARCH
                .getCommandNumber());
        String channelName;
        if (!headerOnly) {
            byte[] channelNameAsBytes = byteSource.getByteArray(messageHeader
                    .getPayloadSize());
            channelName = NullTerminatedStringUtil.nullTerminatedBytesToString(
                    channelNameAsBytes, charset);
        } else {
            channelName = null;
        }
        // The channel CID should be in both the "cid" and "available" fields,
        // however, the server only uses the "available" field.
        return new ChannelAccessSearchClientMessage(messageHeader,
                messageHeader.getContextSpecific(),
                ChannelAccessVersion.forMinorVersion(messageHeader.getCount()),
                messageHeader.getDataType() == DO_REPLY, channelName);
    }

}
